# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow.keras as k

KERNEL_INITIALIZER = k.initializers.RandomNormal(mean=0.0, stddev=0.02)
ALMOST_ONE = k.initializers.RandomNormal(mean=1.0, stddev=0.02)


class Decoder(k.Sequential):
    def __init__(
        self,
        n_channels: int = 3,
        latent_space_dimension: int = 100,
        l2_penalty: float = 0.0,
    ):
        super().__init__(
            [
                k.layers.InputLayer(input_shape=(latent_space_dimension,)),
                k.layers.Dense(512, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.Dense(7 * 7 * 128, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Reshape(target_shape=(7, 7, 128)),
                k.layers.Conv2DTranspose(
                    filters=256,
                    kernel_size=(2, 2),
                    strides=(1, 1),
                    padding="valid",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    filters=128,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    filters=64,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    filters=n_channels,
                    kernel_size=(4, 4),
                    strides=(1, 1),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.Activation(k.activations.tanh),
            ]
        )


class Encoder(k.Sequential):
    def __init__(
        self,
        n_channels: int = 3,
        latent_space_dimension: int = 100,
        l2_penalty: float = 0.0,
    ):
        super().__init__(
            [
                k.layers.InputLayer(input_shape=(32, 32, n_channels)),
                k.layers.Conv2D(
                    filters=64,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=128,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=256,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=latent_space_dimension,
                    kernel_size=(4, 4),
                    strides=(1, 1),
                    padding="valid",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.Flatten(),
            ]
        )


class Discriminator(k.Model):
    def __init__(
        self,
        n_channels: int = 3,
        l2_penalty: float = 0.0,
    ):
        super().__init__()

        # Input
        input_dimension = (32, 32, n_channels)
        self._input_layer = k.layers.InputLayer(input_shape=input_dimension)

        self._backbone = k.Sequential(
            [
                self._input_layer,
                k.layers.Conv2D(
                    filters=32,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=64,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=128,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    filters=256,
                    kernel_size=(4, 4),
                    strides=(2, 2),
                    padding="same",
                    kernel_initializer=KERNEL_INITIALIZER,
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Flatten(),
            ]
        )

        self._output = k.layers.Dense(
            1,
            kernel_initializer=KERNEL_INITIALIZER,
            use_bias=False,
            # kernel_regularizer=k.regularizers.l2(l2_penalty),
        )

    def call(self, inputs, training=True, mask=None):

        features = self._backbone(inputs)
        output = self._output(features)

        return output, features



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DeScarGAN models."""

from abc import abstractmethod
from typing import List

import tensorflow as tf
import tensorflow.keras as k


class DeScarGANModel(k.Model):  # pylint: disable=too-many-ancestors
    """Methods shared across the DeScarGAN models."""

    def __init__(self, batch_norm: bool):
        """Configure the layers.
        Args:
            batch_norm: enable/disable batch normalization in the network description.
        """
        super().__init__()
        self._batch_norm = batch_norm

        self._bias_initializer = None
        self._kernel_initializer = None

    @property
    @abstractmethod
    def kernel_initializer(self):
        pass

    @property
    @abstractmethod
    def bias_initializer(self):
        pass

    def conv(
        self,
        input_shape,
        filters,
        kernel_size=3,
        momentum=0.01,
        activation_layer=k.layers.ReLU(),
    ):
        """Convolutional layer (with or without batch norm depending on the __init__).
        Args:
            input_shape: layer input shape
            filters: number of convolutional filters to learn (the depth of the output volume)
            kernel_size: the size of the kernel to use
            momentum: if batch norm is enabled, the momentum for this layer.
            activation_layer: the activation function to use.
        Returns:
            The convolution operation correctly configured (as a keras model/layer).
        """
        return k.Sequential(
            [
                k.layers.Conv2D(
                    input_shape=input_shape,
                    filters=filters,
                    kernel_size=kernel_size,
                    padding="SAME",
                    kernel_initializer=self.kernel_initializer,
                    bias_initializer=self.bias_initializer,
                    use_bias=not self._batch_norm,
                ),
            ]
            + (
                [k.layers.BatchNormalization(momentum=momentum)]
                if self._batch_norm
                else []
            )
            + [
                activation_layer,
            ]
        )

    @staticmethod
    def deconv(
        input_shape,
        filters,
        use_upsample=True,
        kernel_size=4,
        strides=2,
        padding="SAME",
        momentum=0.01,
        activation_layer=k.layers.ReLU(),
    ):
        """DeConvolutional layer (with or without batch norm depending on the __init__).
        Args:
            input_shape: layer input shape
            filters: number of convolutional filters to learn (the depth of the output volume)
            use_upsample: when True use the upsampling, otherwise conv2d transpose is used.
            kernel_size: the size of the kernel to use
            strides: the stride to use when using conv2d transpose (use_upsample=False)
            padidng: the padding to use when using conv2d transpose (use_upsample=False)
            momentum: if batch norm is enabled, the momentum for this layer.
            activation_layer: the activation function to use.
        Returns:
            The convolution operation correctly configured (as a keras model/layer).
        """
        if use_upsample:
            up_layer = k.Sequential(
                [
                    k.layers.UpSampling2D(),
                    k.layers.Conv2D(
                        input_shape=input_shape,
                        filters=filters,
                        kernel_size=3,
                        strides=1,
                        padding="SAME",
                    ),
                ]
            )
        else:
            up_layer = k.layers.Conv2DTranspose(
                input_shape=input_shape,
                filters=filters,
                kernel_size=kernel_size,
                strides=strides,
                padding=padding,
            )

        return k.Sequential(
            [
                up_layer,
                k.layers.BatchNormalization(momentum=momentum),
                activation_layer,
            ]
        )

    @staticmethod
    def concat(upsampled, bypass):
        """Concatenate bypass and upsampled in the depth dimension.
        Args:
            upsampled: the input batch
            bypass: the skip connection to concatenate
        Return:
            a new tensor with shape (N, H, W, C1 + C2)
        """
        return tf.concat([upsampled, bypass], axis=-1)


class Generator(DeScarGANModel):  # pylint: disable=too-many-ancestors
    """Image generator."""

    def __init__(self, ill_label, n_channels=1, nf=64, batch_norm=True):
        super().__init__(batch_norm)

        self._ill_label = tf.constant(ill_label, dtype=tf.int32)
        self._c_dim = 2
        self._max_pool_fn = k.layers.MaxPool2D

        self._down0 = k.Sequential(
            [
                self.conv((None, None, n_channels + self._c_dim), nf),
                self.conv((None, None, nf), nf),
            ]
        )
        self._down1 = k.Sequential(
            [
                self._max_pool_fn(),
                self.conv((None, None, nf), nf * 2),
                self.conv((None, None, nf * 2), nf * 2),
            ]
        )
        self._down2 = k.Sequential(
            [
                self._max_pool_fn(),
                self.conv((None, None, nf * 2), nf * 4),
                self.conv((None, None, nf * 4), nf * 4),
            ]
        )
        self._down3 = k.Sequential(
            [
                self._max_pool_fn(),
                self.conv((None, None, nf * 4), nf * 8),
                self.conv((None, None, nf * 8), nf * 8),
            ]
        )

        self._up3 = self.deconv((None, None, nf * 8), nf * 4)

        self._conv5 = k.Sequential(
            [
                self.conv((None, None, nf * 8), nf * 4),
                self.conv((None, None, nf * 4), nf * 4),
            ]
        )

        self._up2 = self.deconv((None, None, nf * 4), nf * 2)

        self._conv6 = k.Sequential(
            [
                self.conv((None, None, nf * 4), nf * 2),
                self.conv((None, None, nf * 2), nf * 2),
            ]
        )

        self._up1 = self.deconv((None, None, nf * 2), nf)

        self._conv7_ill = k.Sequential(
            [
                self.conv((None, None, nf), nf),
                self.conv(
                    (None, None, nf),
                    n_channels,
                    activation_layer=k.layers.Activation("tanh"),
                ),
            ]
        )
        self._conv7_healthy = k.Sequential(
            [
                self.conv((None, None, nf), nf),
                self.conv(
                    (None, None, nf),
                    n_channels,
                    activation_layer=k.layers.Activation("tanh"),
                ),
            ]
        )

    @property
    def kernel_initializer(self):
        """kaiming_normal_  in the paper -> HeNormal for keras.
        There are minor differences:
        https://stats.stackexchange.com/questions/484062/he-normal-keras-is-truncated-when-kaiming-normal-pytorch-is-not
        """
        if not self._kernel_initializer:
            self._kernel_initializer = k.initializers.HeNormal()
        return self._kernel_initializer

    @property
    def bias_initializer(self):
        """Zero initializer."""
        if not self._bias_initializer:
            self._bias_initializer = k.initializers.Zeros()
        return self._bias_initializer

    def call(self, inputs: List[tf.Tensor], training=False):
        x, label = inputs
        one_hot_label = tf.one_hot(label, depth=2, dtype=tf.float32)
        condition = tf.tile(
            one_hot_label[:, tf.newaxis, tf.newaxis, :],
            [1, tf.shape(x)[1], tf.shape(x)[2], 1],
        )
        x = tf.concat([x, condition], axis=-1)

        input_0 = self._down0(x, training=training)
        input_1 = self._down1(input_0, training=training)
        input_2 = self._down2(input_1, training=training)
        input_3 = self._down3(input_2, training=training)

        input_upsampled_3 = self._up3(input_3, training=training)
        cat3 = self.concat(input_upsampled_3, input_2)
        input_5 = self._conv5(cat3, training=training)
        input_upsampled_2 = self._up2(input_5, training=training)
        cat2 = self.concat(input_upsampled_2, input_1)

        input_6 = self._conv6(cat2, training=training)
        input_upsampled_1 = self._up1(input_6, training=training)

        return tf.cond(
            tf.reduce_all(tf.equal(self._ill_label, tf.cast(label, tf.int32))),
            lambda: self._conv7_ill(input_upsampled_1),
            lambda: self._conv7_healthy(input_upsampled_1),
        )


class Discriminator(DeScarGANModel):  # pylint: disable=too-many-ancestors
    """Image discriminator."""

    def __init__(self, ill_label, n_channels=1, nf=64, batch_norm=True):
        super().__init__(batch_norm)

        self._ill_label = tf.constant(ill_label, dtype=tf.int32)
        self._max_pool_fn = k.layers.MaxPool2D

        self._encoder = k.Sequential(
            [
                self.conv((None, None, n_channels), nf),
                self._max_pool_fn(),
                self.conv((None, None, nf), nf * 2),
                self._max_pool_fn(),
                self.conv((None, None, nf * 2), nf * 4),
                self.conv((None, None, nf * 4), nf * 4),
                self._max_pool_fn(),
                self.conv((None, None, nf * 4), nf * 8),
                self.conv((None, None, nf * 8), nf * 8),
                self._max_pool_fn(),
                self.conv((None, None, nf * 8), nf * 8),
                self.conv((None, None, nf * 8), nf * 8),
                self._max_pool_fn(),
                self.conv((None, None, nf * 8), nf * 16),
            ]
        )

        self._conv_ill = k.Sequential(
            [
                self.conv((None, None, nf * 16), nf * 16),
                self.conv((None, None, nf * 16), nf * 16),
                self.conv(
                    (None, None, nf * 16),
                    1,
                    kernel_size=1,
                    activation_layer=k.layers.Activation("linear"),
                ),
            ]
        )
        self._conv_healthy = k.Sequential(
            [
                self.conv((None, None, nf * 16), nf * 16),
                self.conv((None, None, nf * 16), nf * 16),
                self.conv(
                    (None, None, nf * 16),
                    1,
                    kernel_size=1,
                    activation_layer=k.layers.Activation("linear"),
                ),
            ]
        )

        self._conv2 = k.Sequential(
            [
                self.conv((None, None, nf * 16), nf * 16),
                self.conv((None, None, nf * 16), nf * 16),
                self._max_pool_fn(),
            ]
        )

        self._linearclass = k.Sequential(
            [
                k.layers.Flatten(),
                k.layers.Dense(64),
                k.layers.ReLU(),
                k.layers.Dropout(0.1),
                k.layers.Dense(2),
            ]
        )

    @property
    def kernel_initializer(self):
        """Xavier normal"""
        if not self._kernel_initializer:
            self._kernel_initializer = k.initializers.GlorotNormal()
        return self._kernel_initializer

    @property
    def bias_initializer(self):
        """Zero initializer."""
        if not self._bias_initializer:
            self._bias_initializer = k.initializers.Zeros()
        return self._bias_initializer

    def call(self, inputs: List[tf.Tensor], training=False):
        x, label = inputs
        hidden = self._encoder(x, training=training)
        out = tf.cond(
            tf.reduce_all(tf.equal(self._ill_label, tf.cast(label, tf.int32))),
            lambda: self._conv_ill(hidden, training=training),
            lambda: self._conv_healthy(hidden, training=training),
        )
        conv = self._conv2(hidden, training=training)
        pred = self._linearclass(tf.squeeze(conv, axis=[1, 2]), training=training)
        return out, pred



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Anomaly detection dataset interface."""

import abc
from typing import Tuple, Union

import tensorflow as tf


class AnomalyDetectionDataset(abc.ABC):
    """
    Anomaly detection dataset interface.
    Note:
        The label 1 is for the ANOMALOUS class.
        The label 0 is for the NORMAL class.
    """

    def __init__(self):
        self._train = None
        self._train_normal = None
        self._train_anomalous = None

        self._test = None
        self._test_normal = None
        self._test_anomalous = None

        self._validation = None
        self._validation_normal = None
        self._validation_anomalous = None

        self._channels = 1

        self._anomalous_label = tf.constant(1)
        self._normal_label = tf.constant(0)

    @staticmethod
    def linear_conversion(
        image: tf.Tensor, new_min: tf.Tensor, new_max: tf.Tensor
    ) -> tf.Tensor:
        """Linearly convert image from it's actual value range,
        to a new range (new_min, new_max).
        Useful to change the values of an image in a new range.

        Args:
            image: float tensor
            new_min: float value or tensor
            new_max: new maximum value, float value or tensor
        Returns:
            a tensor with the same shape and type of image, but
            with values in  [new_min, new_max]
        """
        old_min = tf.reduce_min(image)
        old_max = tf.reduce_max(image)
        return (image - old_min) * (new_max - new_min) / (old_max - old_min) + new_min

    @property
    def channels(self) -> int:
        """The last dimension of the elements in the dataset.
        e.g. 3 if the dataset is a dataset of RGB images or 1
        if they are grayscale."""
        return self._channels

    @property
    def anomalous_label(self) -> tf.Tensor:
        """Return the constant tensor used for anomalous label (1)."""
        return self._anomalous_label

    @property
    def normal_label(self) -> tf.Tensor:
        """Return the constant tensor used for normal data label (0)."""
        return self._normal_label

    @property
    def train_normal(self) -> tf.data.Dataset:
        """Subset of the training dataset: only positive."""
        return self._train_normal

    @property
    def train_anomalous(self) -> tf.data.Dataset:
        """Subset of the training dataset: only negative."""
        return self._train_anomalous

    @property
    def train(self) -> tf.data.Dataset:
        """The complete training dataset with both positive and negatives.
        The labels are always 2."""
        return self._train

    @property
    def test(self) -> tf.data.Dataset:
        """The complete test dataset with both positive and negatives.
        The labels are always 2."""
        return self._test

    @property
    def test_normal(self) -> tf.data.Dataset:
        """Subset of the test dataset: only positive."""
        return self._test_normal

    @property
    def test_anomalous(self) -> tf.data.Dataset:
        """Subset of the test dataset: only negative."""
        return self._test_anomalous

    @property
    def validation(self) -> tf.data.Dataset:
        """The complete test dataset with both positive and negatives.
        The labels are always 2."""
        return self._validation

    @property
    def validation_normal(self) -> tf.data.Dataset:
        """Subset of the validation dataset: only positive."""
        return self._validation_normal

    @property
    def validation_anomalous(self) -> tf.data.Dataset:
        """Subset of the validation dataset: only negative."""
        return self._validation_anomalous

    @abc.abstractmethod
    def configure(
        self,
        batch_size: int,
        new_size: Tuple[int, int],
        anomalous_label: Union[int, str, None] = None,
        class_label: Union[int, str, None] = None,
        shuffle_buffer_size: int = 10000,
        cache: bool = True,
        drop_remainder: bool = True,
        output_range: Tuple[float, float] = (-1.0, 1.0),
    ) -> None:
        """Configure the dataset. This makes all the object properties valid (not None).
        Args:
            batch_size: The dataset batch size
            new_size: (H,W) of the input image.
            anomalous_label: If the raw dataset contains labels, all the elements with
                             "anomalous_label" are converted to element of
                             self.anomalous_label class.
            class_label: If the raw dataset contains different classes (each one
                         containing both positive and negative samples) we can select
                         only one class to focus on (e.g. a dataset of industrial
                         defects on industrial objects composed of transistors and
                         pills and we are interested only in transistors and not on pills).
            shuffle_buffer_size: Buffer size used during the tf.data.Dataset.shuffle call.
            cache: If True, cache the dataset
            drop_remainder: If True, when the dataset size is not a multiple of the dataset size,
                            the last batch will be dropped.
            output_range: A Tuple (min, max) containing the output range to use for
                          the processed images.
        """
        raise NotImplementedError

    @staticmethod
    def pipeline(
        dataset: tf.data.Dataset,
        new_size: Tuple[int, int],
        batch_size: int,
        cache: bool,
        shuffle_buffer_size: int,
        is_training: bool = True,
        drop_remainder: bool = True,
        output_range: Tuple[float, float] = (0.0, 1.0),
    ) -> tf.data.Dataset:
        """Given a dataset, configure it applying the chain of
        map, filter, shuffle and all the needed methods of the tf.data.Dataset.
        Args:
            dataset: the input dataset with elements (x,y), where x is an image
                     and y the scalar label.
            new_size: (H,W) of the output image. NEAREST_NEIGHBOR interpolation is used.
            batch_size: the dataset batch size
            cache: when true, calls the `.cache()` method on the dataset before
                   returning it.
            shuffle_buffer_size: buffer size used during the tf.data.Dataset.shuffle call.
            is_training: when true, shuffles the dataset element using a shuffle buffer with
                         size shuffle_buffer_size.
            drop_remainder: if True, when the dataset size is not a multiple of the dataset size,
                            the last batch will be dropped.
            output_range: a Tuple (min, max) containing the output range to use
                          for the processed images.
        Returns:
            The configured dataset object.
        """
        dataset = dataset.map(
            lambda image, label: (
                tf.image.resize(
                    image, size=new_size, method=tf.image.ResizeMethod.NEAREST_NEIGHBOR
                ),
                label,
            )
        )

        dataset = dataset.map(
            lambda image, label: (tf.cast(image, tf.float32) / 255.0, label)
        )

        if output_range != (0.0, 1.0):
            dataset = dataset.map(
                lambda image, label: (
                    AnomalyDetectionDataset.linear_conversion(
                        image, output_range[0], output_range[1]
                    ),
                    label,
                )
            )

        if is_training:
            dataset = dataset.shuffle(buffer_size=shuffle_buffer_size)
        dataset = dataset.batch(batch_size, drop_remainder=drop_remainder)
        if cache:
            dataset = dataset.cache()
        return dataset



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import tensorflow as tf
from tensorflow import keras


class AdversarialLoss(keras.losses.Loss):
    """
    Adversarial loss.
    """

    def __init__(self, from_logits: bool = True):
        super().__init__()
        self._bce = keras.losses.BinaryCrossentropy(from_logits=from_logits)

    def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor):
        d_real = y_true
        d_gen = y_pred
        real_loss = self._bce(tf.ones_like(d_real), d_real)
        generated_loss = self._bce(tf.zeros_like(d_gen), d_gen)
        return real_loss + generated_loss


def generator_bce(d_gz: tf.Tensor, from_logits: bool = True):
    """
    Calculated the binary cross entropy loss of the generator.

    Args:
        d_gz: Discriminator classification on the generator reconstructed data.
        from_logits: True if the values are unbounded, False if they are a probability
        distribution [0, 1].

    Returns:
        The result of the keras.losses.BinaryCrossentropy function.
    """
    return keras.losses.BinaryCrossentropy(from_logits=from_logits)(
        tf.ones_like(d_gz), d_gz
    )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Surface Crack Dataset (https://www.kaggle.com/arunrk7/surface-crack-detection).

The datasets contains images of various concrete surfaces with and without crack.
The image data are divided into two as negative (without crack) and positive
(with crack) in separate folder for image classification.
Each class has 20000 images with a total of 40000 images with 227 x 227 pixels
with RGB channels.
The dataset is generated from 458 high-resolution images (4032x3024 pixel) with
the method proposed by Zhang et al (2016).
High resolution images found out to have high variance in terms of surface finish
and illumination condition.
No data augmentation in terms of random rotation or flipping or tilting is applied.
"""

import zipfile
from functools import partial
from glob import glob
from io import BytesIO
from pathlib import Path
from typing import Tuple, Union

import rarfile
import requests
import tensorflow as tf

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset


class SurfaceCracks(AnomalyDetectionDataset):
    """Surface Crack Dataset (https://www.kaggle.com/arunrk7/surface-crack-detection).
    20000 images for training (balanced), 5000 images for validation (balanced)
    5000 images for testing (balanced).
    """

    def __init__(self, path: Path = Path("surface_cracks")):
        super().__init__()

        self._archive_url = (
            "https://"
            "md-datasets-cache-zipfiles-prod.s3.eu-west-1"
            ".amazonaws.com/5y9wdsg2zt-2.zip"
        )
        self._path = path

        self._download_and_extract()

        def _read_and_map_fn(label):
            """Closure used in tf.data.Dataset.map for creating
            the correct pair (image, label).
            """

            def fn(filename):
                binary = tf.io.read_file(filename)
                image = tf.image.decode_jpeg(binary)
                return image, label

            return fn

        glob_ext = "*.jpg"
        all_normal = glob(str(self._path / "Negative" / glob_ext))
        all_normal_train = all_normal[:10000]
        all_normal_test = all_normal[10000:15000]
        all_normal_validation = all_normal[15000:]

        all_anomalous = glob(str(self._path / "Positive" / glob_ext))
        all_anomalous_train = all_anomalous[:10000]
        all_anomalous_test = all_anomalous[10000:15000]
        all_anomalous_validation = all_anomalous[15000:]

        self._train_raw = (
            tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(all_normal_train))
            .map(_read_and_map_fn(self.normal_label))
            .concatenate(
                tf.data.Dataset.from_tensor_slices(
                    tf.convert_to_tensor(all_anomalous_train)
                ).map(_read_and_map_fn(self.anomalous_label))
            )
        )

        self._test_raw = (
            tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(all_normal_test))
            .map(_read_and_map_fn(self.normal_label))
            .concatenate(
                tf.data.Dataset.from_tensor_slices(
                    tf.convert_to_tensor(all_anomalous_test)
                ).map(_read_and_map_fn(self.anomalous_label))
            )
        )

        self._validation_raw = (
            tf.data.Dataset.from_tensor_slices(
                tf.convert_to_tensor(all_normal_validation)
            )
            .map(_read_and_map_fn(self.normal_label))
            .concatenate(
                tf.data.Dataset.from_tensor_slices(
                    tf.convert_to_tensor(all_anomalous_validation)
                ).map(_read_and_map_fn(self.anomalous_label))
            )
        )

        # RGB dataset
        self._channels = 3

    def _download_and_extract(self) -> None:
        """Download and extract the dataset."""

        if self._path.exists():
            print(self._path, " already exists. Skipping dataset download.")
            return
        self._path.mkdir()

        # Download a zip file
        print("Downloading dataset from: ", self._archive_url)
        request = requests.get(self._archive_url)
        print("Unzipping...")
        with zipfile.ZipFile(BytesIO(request.content)) as zip_archive:
            # The zip file contains a rar file :\
            print(
                "Unrarring... "
                "(this may take up to 15 minutes because python 'rarfile' is slow)"
            )
            rar_archive = zip_archive.read(
                "Concrete Crack Images for Classification.rar"
            )
            rar_path = self._path / "cracks.rar"
            with open(str(rar_path), "wb") as fp:
                fp.write(rar_archive)

            rar = rarfile.RarFile(str(rar_path))
            rar.extractall(str(self._path))
        print("Raw dataset downloaded and extracted.")

    def configure(
        self,
        batch_size: int,
        new_size: Tuple[int, int],
        anomalous_label: Union[int, str, None] = None,
        class_label: Union[int, str, None] = None,
        shuffle_buffer_size: int = 10000,
        cache: bool = True,
        drop_remainder: bool = True,
        output_range: Tuple[float, float] = (0.0, 1.0),
    ) -> None:
        """Configure the dataset. This makes all the object properties valid (not None).
        Args:
            batch_size: The dataset batch size
            new_size: (H,W) of the input image.
            anomalous_label: If the raw dataset contains label, all the elements with
                             "anomalous_label" are converted to element of
                             self.anomalous_label class.
            class_label: If the raw dataset contains different classes (each one
                         containing both positive and negative samples) we can select
                         only one class to focus on (e.g. a dataset of industrial
                         defects on industrial objects composed of transistors and
                         pills and we are interested only in transistors and not on pills).
            shuffle_buffer_size: Buffer size used during the tf.data.Dataset.shuffle call.
            cache: If True, cache the dataset
            drop_remainder: If True, when the dataset size is not a multiple of the dataset size,
                            the last batch will be dropped.
            output_range: A Tuple (min, max) containing the output range to use
                          for the processed images.
        """

        pipeline = partial(
            self.pipeline,
            new_size=new_size,
            batch_size=batch_size,
            shuffle_buffer_size=shuffle_buffer_size,
            cache=cache,
            drop_remainder=drop_remainder,
            output_range=output_range,
        )

        pipeline_train = partial(pipeline, is_training=True)
        pipeline_test = partial(pipeline, is_training=False)
        is_anomalous = lambda _, label: tf.equal(label, anomalous_label)
        is_normal = lambda _, label: tf.not_equal(label, anomalous_label)

        # Train-data
        self._train_anomalous = (
            self._train_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_train)
        )
        self._train_normal = (
            self._train_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_train)
        )
        self._train = self._train_raw.apply(pipeline_train)

        # Test-data
        self._validation_anomalous = (
            self._validation_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_test)
        )
        self._validation_normal = (
            self._validation_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_test)
        )
        self._validation = self._validation_raw.apply(pipeline_test)

        # Test-data
        self._test_anomalous = (
            self._test_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_test)
        )
        self._test_normal = (
            self._test_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_test)
        )
        self._test = self._test_raw.apply(pipeline_test)



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
BiGAN losses according to EGBAD.

The losses here defined are the standard GAN losses (non saturating + sum BCEs).
"""

import tensorflow as tf
from tensorflow import keras


class AdversarialLoss(keras.losses.Loss):
    """
    Adversarial loss.
    """

    def __init__(self, from_logits: bool = True):
        super().__init__()
        self._bce = keras.losses.BinaryCrossentropy(from_logits=from_logits)

    def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor):
        d_real = y_true
        d_gen = y_pred
        real_loss = self._bce(tf.ones_like(d_real), d_real)
        generated_loss = self._bce(tf.zeros_like(d_gen), d_gen)
        return real_loss + generated_loss


def residual_loss(x: tf.Tensor, g_z: tf.Tensor, axis: int = -1):
    """
    Return the residual loss between x and Gz.

    Args:
        x: The original images batch, 4D.
        g_z: The generated images.
        axis: The axis on which perform the reduce operation, default: -1.

    Returns:
        sum | Gz - x |
    """
    assert x.shape == g_z.shape
    flat = (-1, x.shape[1] * x.shape[2] * x.shape[3])
    return tf.reduce_mean(tf.reshape(tf.abs(g_z - x), shape=flat), axis=axis)


def generator_bce(d_gz: tf.Tensor, from_logits: bool = True):
    """
    Calculated the binary cross entropy loss of the generator.

    Args:
        d_gz: Discriminator classification on the generator reconstructed data.
        from_logits: True if the values are unbounded, False if they are a probability
        distribution [0, 1].

    Returns:
        The result of the keras.losses.BinaryCrossentropy function.
    """
    return keras.losses.BinaryCrossentropy(from_logits=from_logits)(
        tf.ones_like(d_gz), d_gz
    )


def encoder_bce(d_x: tf.Tensor, from_logits: bool = True):
    """
    Calculated the binary cross entropy loss of the encoder.

    Args:
        d_x: Discriminator classification on the real data.
        from_logits: True if the values are unbounded, False if they are a probability
        distribution [0, 1].

    Returns:
        The result of the keras.losses.BinaryCrossentropy function.
    """
    return keras.losses.BinaryCrossentropy(from_logits=from_logits)(
        tf.zeros_like(d_x), d_x
    )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Union

import tensorflow as tf
import tensorflow.keras as keras

from anomaly_toolbox.models.ganomaly import GANomalyGenerator


class GANomalyPredictor:
    generator: keras.Model
    discriminator: keras.Model

    def load_from_savedmodel(self, generator_dir: str, discriminator_dir: str):
        self.generator: GANomalyGenerator = tf.keras.models.load_model(generator_dir)
        self.discriminator = tf.keras.models.load_model(discriminator_dir)

    def evaluate(self, dataset: tf.data.Dataset) -> Tuple[tf.Tensor, tf.Tensor]:
        """Evaluate for benchmark."""
        anomaly_scores, labels = [], []
        for batch in dataset:
            # a_score: [batch, 1, latent dimension]
            # y: [batch, 1]
            a_score, y = self.evaluate_step(batch)
            anomaly_scores.append(a_score)
            labels.append(y)
        anomaly_scores = tf.reshape(anomaly_scores, -1)
        labels = tf.reshape(labels, -1)
        tf.assert_equal(anomaly_scores.shape, labels.shape)
        return anomaly_scores, labels

    @tf.function
    def evaluate_step(
        self, inputs: Tuple[tf.Tensor, tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        # x: [batch, height, width, channels]
        # y: [batch, 1]
        x, y = inputs

        # z: [batch, 1, 1, latent dimension]
        # x_hat: [batch, height, width, channels]
        # z_hat: [batch, 1, 1, latent dimension]
        z, x_hat, z_hat = self.generator(x)

        # z: [batch, latent dimension]
        # z_hat: [batch, latent dimension]
        z, z_hat = tf.squeeze(z), tf.squeeze(z_hat)

        # a_score: [batch, 1]
        a_score = self.compute_anomaly_score(z, z_hat)

        return a_score, y

    @staticmethod
    @tf.function
    def predict(
        generator: keras.Model, x: tf.Tensor, return_score_only: bool = True
    ) -> Union[tf.Tensor, Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]:
        # x: [batch, height, width, channels]

        # z: [batch, 1, 1, latent dimension]
        # x_hat: [batch, height, width, channels]
        # z_hat: [batch, 1, 1, latent dimension]
        z, x_hat, z_hat = generator(x)

        # z: [batch, latent dimension]
        # z_hat: [batch, latent dimension]
        z, z_hat = tf.squeeze(z), tf.squeeze(z_hat)

        # a_score: [batch, 1]
        a_score = GANomalyPredictor.compute_anomaly_score(z, z_hat)
        if return_score_only:
            return a_score
        else:
            return x_hat, z_hat, z, a_score

    @staticmethod
    def compute_anomaly_score(
        encoded_input: tf.Tensor, encoded_generated: tf.Tensor
    ) -> tf.Tensor:
        anomaly_score = tf.reduce_mean(
            tf.math.squared_difference(encoded_input, encoded_generated), axis=1
        )
        return anomaly_score



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Experiment generic structure."""

import abc
from pathlib import Path
from typing import Callable, Dict, List, Set

from tensorboard.plugins.hparams import api as hp

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset


class Experiment(abc.ABC):
    """
    The Experiment class represent the basic class to be used by each experiment.
    """

    def __init__(self, hparams_file_path: Path, log_dir: Path):
        """
        Experiment ctor.

        Args:
            hparams_file_path: The path of the hparams JSON file.
            log_dir: The log dir to use during the experiment.
        """
        self._hparams_path = hparams_file_path
        self._log_dir = log_dir
        self._hps = []

    @property
    def log_dir(self) -> Path:
        """The log dir to use during the experiment."""
        return self._log_dir

    @property
    def hps(self) -> List[hp.HParam]:
        """The list of the hyperparameters."""
        return self._hps

    @staticmethod
    def hyperparameters() -> Set[str]:
        """Common hyperparameters to all the experiments. These are the
        data-related hyperparameters."""
        return {
            "anomalous_label",
            "class_label",
            "epochs",
            "batch_size",
            "shuffle_buffer_size",
            "step_log_frequency",
        }

    @abc.abstractmethod
    def experiment(
        self, hps: Dict, log_dir: Path, dataset: AnomalyDetectionDataset
    ) -> None:
        """Experiment execution - architecture specific.
        Args:
            hps: Dictionary with the parameters to use for the current run.
            log_dir: Where to store the tensorboard logs.
            dataset: The datset to use for model training and evaluation.
        """
        raise NotImplementedError

    def run(
        self,
        hparams_tuning: bool,
        hparams_func: Callable,
        dataset: AnomalyDetectionDataset,
    ) -> None:
        """
        Run training. It can run with hyperparameters tuning (hparams_tuning==True) or it can
        run without tuning (hparams_tuning==False). The tuning function is passed as a callable.

        Args:
            hparams_tuning: True if you want to enable tuning, False otherwise.
            hparams_func: The tuning function (e.g., grid_search) to use to do hyperparameters
            tuning.
            dataset: The datset to use for model training and evaluation.
        """
        if hparams_tuning:
            hparams_func(
                self.experiment,
                hps=self.hps,
                log_dir=self.log_dir,
                dataset=dataset,
            )
        else:
            hps_run = {entry.name: entry.domain.values[0] for entry in self.hps}
            self.experiment(hps_run, self.log_dir, dataset)



#!/usr/bin/env python3

"""Training & evaluation script for anomaly toolbox."""

# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import json
import logging
import os
import sys
from pathlib import Path

import click
from tabulate import tabulate

import anomaly_toolbox.datasets as available_datasets
import anomaly_toolbox.experiments as available_experiments
from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.hps import grid_search


@click.command()
@click.option(
    "chosen_experiment",
    "--experiment",
    help="Experiment to run.",
    type=click.Choice(available_experiments.__experiments__, case_sensitive=True),
)
@click.option(
    "hps_path",
    "--hps-path",
    help="When running an experiment, the path of the JSON file where all "
    "the hyperparameters are located.",
    type=Path,
    required=True,
)
@click.option(
    "hps_tuning",
    "--tuning",
    help="If you want to use hyperparameters tuning, use 'True' here. Default is False.",
    type=bool,
    default=False,
)
@click.option(
    "dataset",
    "--dataset",
    help=(
        "The dataset to use. Can be a ready to use dataset, or a .py file "
        "that implements the AnomalyDetectionDataset interface"
    ),
    type=str,
    required=True,
)
@click.option(
    "run_all",
    "--run-all/--no-run-all",
    help="Run all the available experiments",
    type=bool,
    default=False,
)
def main(
    chosen_experiment: str,
    hps_path: Path,
    hps_tuning: bool,
    dataset: str,
    run_all: bool,
) -> int:
    """CLI anomaly-box."""

    # Warning to the user if the hparams_tuning.json && --tuning==False
    if "tuning" in str(hps_path) and not hps_tuning:
        logging.warning(
            "You choose to use the tuning JSON but the tuning boolean ("
            "--tuning) is False. Only one kind of each parameters will be taken "
            "into consideration. No tuning will be performed."
        )

    # Instantiate dataset config from dataset name
    if dataset.endswith(".py"):
        file_path = Path(dataset).absolute()
        name = file_path.stem
        sys.path.append(str(file_path.parent))

        dataset_instance = getattr(__import__(name), name)()
        if not isinstance(dataset_instance, AnomalyDetectionDataset):
            logging.error(
                "Your class %s must implement the "
                "anomaly_toolbox.datasets.dataset.AnomalyDetectionDataset"
                "interface",
                dataset_instance,
            )
            return 1
    else:
        try:
            dataset_instance = getattr(
                importlib.import_module("anomaly_toolbox.datasets"),
                dataset,
            )()
        except (ModuleNotFoundError, AttributeError, TypeError) as exc:
            logging.error(
                "Dataset %s is not among the available: %s.\n\nRaw error: %s",
                dataset,
                ",".join(available_datasets.__datasets__),
                exc,
            )
            return 1

    if run_all and chosen_experiment:
        logging.error("Only one between --run_all and --experiment can be used.")
        return 1

    if not (chosen_experiment or run_all):
        logging.error(
            "Please choose a valid CLI flag.\n%s --help", Path(sys.argv[0]).name
        )
        return 1

    if not hps_path or not hps_path.exists():
        logging.error(
            "Check that %s exists and it's a valid JSON containing the hyperparameters.",
            hps_path,
        )
        return 1

    if chosen_experiment:
        experiments = [chosen_experiment]
    else:
        experiments = available_experiments.__experiments__

    for experiment in experiments:
        log_dir = Path("logs") / experiment
        log_dir.mkdir(parents=True, exist_ok=True)

        try:
            experiment_instance = getattr(
                importlib.import_module("anomaly_toolbox.experiments"),
                experiment,
            )(hps_path, log_dir)
        except (ModuleNotFoundError, AttributeError, TypeError) as exc:
            logging.error(
                "Experiment %s is not among the available: %s.\n\nRaw error: %s",
                experiment,
                ",".join(available_experiments.__experiments__),
                exc,
            )
            return 1

        results = log_dir / "results"
        if results.exists():
            print("Skipping: ", experiment, " result folder already present")
            continue

        experiment_instance.run(hps_tuning, grid_search, dataset_instance)

    # Check the best result for all experiments
    if run_all:
        result_dirs = ["auc_rc", "auc_roc", "accuracy", "precision", "recall", "f1"]

        # For every experiment
        table = [result_dirs]  # This create the table header
        for experiment in experiments:
            table_row = [[experiment] + [0.0] * len(result_dirs)]
            log_dir = Path("logs") / experiment / "results"
            if os.path.exists(log_dir):
                dirs = os.listdir(log_dir)

                # For every possible metric collected
                for metric in [
                    directory for directory in dirs if directory in result_dirs
                ]:
                    json_file = log_dir / metric / "test.json"

                    # Get the value from the json file
                    with open(json_file, "r") as file:
                        data = json.load(file)

                    if metric in data:
                        # Put the result in the correct position of the list
                        # Get the index inside the list
                        idx = result_dirs.index(metric) + 1
                        table_row[0][idx] = data[metric]["value"]

                # Put the results in the table
                table = table + table_row

        md_table = tabulate(table, headers="firstrow", tablefmt="github")
        with open("result_table.md", "w") as outputfile:
            outputfile.write(md_table)

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DeScarGAN experiment suite."""

from pathlib import Path
from typing import Dict

import tensorflow as tf

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.experiments.experiment import Experiment
from anomaly_toolbox.hps import hparam_parser
from anomaly_toolbox.trainers import DeScarGAN


class DeScarGANExperiment(Experiment):
    """
    DeScarGAN experiment.
    """

    def __init__(self, hparams_path: Path, log_dir: Path):
        super().__init__(hparams_path, log_dir)

        # Get the hyperparameters
        self._hps = hparam_parser(
            hparams_path,
            "descargan",
            list(self.hyperparameters().union(DeScarGAN.hyperparameters())),
        )

    def experiment(
        self, hps: Dict, log_dir: Path, dataset: AnomalyDetectionDataset
    ) -> None:
        """Experiment execution - architecture specific.
        Args:
            hps: Dictionary with the parameters to use for the current run.
            log_dir: Where to store the tensorboard logs.
            dataset: The dataset to use for model training and evaluation.
        """

        print("Running DeScarGAN experiment...")

        # Create and configure the dataset
        dataset.configure(
            anomalous_label=hps["anomalous_label"],
            class_label=hps["class_label"],
            batch_size=hps["batch_size"],
            new_size=(64, 64),
            output_range=(-1.0, 1.0),  # generator has a tanh in output
            shuffle_buffer_size=hps["shuffle_buffer_size"],
            cache=True,
        )

        summary_writer = tf.summary.create_file_writer(str(log_dir))

        # Create the DeScarGAN model
        trainer = DeScarGAN(dataset, hps, summary_writer, log_dir)

        # Train the model
        trainer.train(
            epochs=hps["epochs"],
            step_log_frequency=hps["step_log_frequency"],
        )

        # Test the model
        trainer.test()



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the 28x28 input resolution models of AnoGAN."""
from typing import Tuple

import tensorflow as tf
import tensorflow.keras as k


class Generator(k.Sequential):
    """Generator in fully convolutional fashion.
    Input: 1x1x input_dimension.
    """

    def __init__(self, n_channels: int = 3, input_dimension: int = 128):
        """Generator model.
        Args:
            n_channels: Depth of the input image.
            input_dimension: The dimension of the latent vector.
        """
        super().__init__(
            [
                k.layers.InputLayer(input_shape=(input_dimension,)),
                k.layers.Dense(7 * 7 * 128, use_bias=False),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Reshape((7, 7, 128)),
                k.layers.Conv2DTranspose(
                    128, (2, 2), strides=(1, 1), use_bias=False, padding="same"
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    64, (2, 2), strides=(2, 2), use_bias=False, padding="same"
                ),
                k.layers.BatchNormalization(),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    n_channels, (5, 5), strides=(2, 2), use_bias=False, padding="same"
                ),
                k.layers.Activation("tanh"),
            ]
        )


class Discriminator(k.Model):
    """Discriminator model. Expects a batch of 28x28x1 input images."""

    def __init__(self, n_channels: int = 3):
        """
        Args:
            n_channels: Depth of the input image.
        """
        super().__init__()

        self._features = k.Sequential(
            [
                k.layers.InputLayer(input_shape=(28, 28, n_channels)),
                k.layers.Conv2D(32, (5, 5), padding="same", strides=(2, 2)),
                k.layers.LeakyReLU(0.2),
                k.layers.Dropout(0.5),
            ]
        )

        self._classifier = k.Sequential(
            [
                k.layers.InputLayer(input_shape=(14, 14, 32)),
                k.layers.Conv2D(64, (5, 5), strides=(2, 2), padding="same"),
                k.layers.LeakyReLU(0.2),
                k.layers.Dropout(0.5),
                k.layers.Flatten(),
                k.layers.Dense(1),
            ]
        )

    def call(
        self, inputs: tf.Tensor, training: bool = False
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """Forward pass.
        Args:
            inputs: Input batch.
            training: Toggle the model status from training to inference.
        Returns:
            out, features.

            out: The discriminator decision (single neuron, linear activation).
            features: The feature vector computed.
        """

        features = self._features(inputs, training=training)
        out = self._classifier(features, training=training)
        return out, features



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trainer for the GANomaly model."""

import json
from pathlib import Path
from typing import Dict, Set, Union

import tensorflow as tf
import tensorflow.keras as k

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.losses.ganomaly import AdversarialLoss, generator_bce
from anomaly_toolbox.models.ganomaly import Decoder, Discriminator, Encoder
from anomaly_toolbox.trainers.trainer import Trainer


class GANomaly(Trainer):
    """GANomaly Trainer."""

    def __init__(
        self,
        dataset: AnomalyDetectionDataset,
        hps: Dict,
        summary_writer: tf.summary.SummaryWriter,
        log_dir: Path,
    ):
        """Initialize GANomaly Networks."""
        super().__init__(
            dataset=dataset, hps=hps, summary_writer=summary_writer, log_dir=log_dir
        )

        n_channels = dataset.channels

        # Discriminator
        self.discriminator = Discriminator(n_channels=n_channels, l2_penalty=0.2)

        # Generator (aka Decoder)
        self.generator = Decoder(
            n_channels=n_channels,
            latent_space_dimension=self._hps["latent_vector_size"],
            l2_penalty=0.2,
        )

        # Encoder
        self.encoder = Encoder(
            n_channels=n_channels,
            latent_space_dimension=self._hps["latent_vector_size"],
            l2_penalty=0.2,
        )

        fake_batch_size = (1, 32, 32, n_channels)
        self.discriminator(tf.zeros(fake_batch_size))
        self.discriminator.summary()

        self.encoder(tf.zeros(fake_batch_size))
        self.encoder.summary()

        fake_latent_vector = (1, self._hps["latent_vector_size"])
        self.generator(tf.zeros(fake_latent_vector))
        self.generator.summary()

        # Losses
        self._mse = k.losses.MeanSquaredError()
        self._mae = k.losses.MeanAbsoluteError()

        # Optimizers
        self.optimizer_ge = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.optimizer_d = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )

        # Training Metrics
        self.epoch_g_loss_avg = k.metrics.Mean(name="epoch_generator_loss")
        self.epoch_d_loss_avg = k.metrics.Mean(name="epoch_discriminator_loss")
        self.epoch_e_loss_avg = k.metrics.Mean(name="epoch_encoder_loss")

        self._auc_rc = k.metrics.AUC(name="auc_rc", curve="PR", num_thresholds=500)
        self._auc_roc = k.metrics.AUC(name="auc_roc", curve="ROC", num_thresholds=500)

        self.keras_metrics = {
            metric.name: metric
            for metric in [
                self.epoch_d_loss_avg,
                self.epoch_g_loss_avg,
                self.epoch_e_loss_avg,
                self._auc_rc,
                self._auc_roc,
            ]
        }

        self._minmax = AdversarialLoss(from_logits=True)

        # Flatten op
        self._flatten = k.layers.Flatten()

    @staticmethod
    def hyperparameters() -> Set[str]:
        """List of the hyperparameters name used by the trainer."""
        return {
            "learning_rate",
            "latent_vector_size",
            "adversarial_loss_weight",
            "contextual_loss_weight",
            "enc_loss_weight",
        }

    def train(
        self,
        epochs: int,
        adversarial_loss_weight: float,
        contextual_loss_weight: float,
        enc_loss_weight: float,
        step_log_frequency: int = 100,
    ):
        best_auc_rc, best_auc_roc = -1, -1

        for epoch in range(epochs):
            for batch in self._dataset.train_normal:
                x, _ = batch

                # Perform the train step
                g_z, g_ex, d_loss, g_loss, e_loss = self.train_step(
                    x,
                    adversarial_loss_weight,
                    contextual_loss_weight,
                    enc_loss_weight,
                )

                # Update the losses metrics
                self.epoch_g_loss_avg.update_state(g_loss)
                self.epoch_d_loss_avg.update_state(d_loss)
                self.epoch_e_loss_avg.update_state(e_loss)

                step = self.optimizer_d.iterations.numpy()
                learning_rate = self.optimizer_ge.learning_rate.numpy()

                if tf.equal(tf.math.mod(step, step_log_frequency), 0):
                    with self._summary_writer.as_default():
                        tf.summary.scalar("learning_rate", learning_rate, step=step)
                        tf.summary.image(
                            "x/g_z/g_ex",
                            tf.concat([x, g_z, g_ex], axis=2),
                            step=step,
                        )
                        tf.summary.scalar(
                            "d_loss",
                            self.epoch_d_loss_avg.result(),
                            step=step,
                        )
                        tf.summary.scalar(
                            "g_loss",
                            self.epoch_g_loss_avg.result(),
                            step=step,
                        )
                        tf.summary.scalar(
                            "e_loss",
                            self.epoch_e_loss_avg.result(),
                            step=step,
                        )

                    tf.print(
                        "Step ",
                        step,
                        ". d_loss: ",
                        self.epoch_d_loss_avg.result(),
                        ", g_loss: ",
                        self.epoch_g_loss_avg.result(),
                        ", e_loss: ",
                        self.epoch_e_loss_avg.result(),
                        "lr: ",
                        learning_rate,
                    )

            # Epoch end
            tf.print(epoch, "Epoch completed")

            # Model selection
            self._auc_rc.reset_state()
            self._auc_roc.reset_state()
            for batch in self._dataset.validation:
                x, labels_test = batch

                anomaly_scores = self._compute_anomaly_scores(
                    x, self.encoder, self.generator
                )
                self._auc_rc.update_state(labels_test, anomaly_scores)
                self._auc_roc.update_state(labels_test, anomaly_scores)

            # Save the model when AUC-RC is the best
            current_auc_rc = self._auc_rc.result()
            if best_auc_rc < current_auc_rc:
                tf.print("Best AUC-RC on validation set: ", current_auc_rc)

                # Replace the best
                best_auc_rc = current_auc_rc

                base_path = self._log_dir / "results" / "auc_rc"
                self.generator.save(str(base_path / "generator"), overwrite=True)
                self.encoder.save(str(base_path / "encoder"), overwrite=True)
                self.discriminator.save(
                    str(base_path / "discriminator"), overwrite=True
                )

                with open(base_path / "validation.json", "w") as fp:
                    json.dump(
                        {
                            "value": float(best_auc_rc),
                        },
                        fp,
                    )
            # Save the model when AUC-ROC is the best
            current_auc_roc = self._auc_roc.result()
            if best_auc_roc < current_auc_roc:
                tf.print("Best AUC-ROC on validation set: ", current_auc_roc)

                # Replace the best
                best_auc_roc = current_auc_roc

                base_path = self._log_dir / "results" / "auc_roc"
                self.generator.save(str(base_path / "generator"), overwrite=True)
                self.encoder.save(str(base_path / "encoder"), overwrite=True)
                self.discriminator.save(
                    str(base_path / "discriminator"), overwrite=True
                )

                with open(base_path / "validation.json", "w") as fp:
                    json.dump(
                        {
                            "value": float(best_auc_rc),
                        },
                        fp,
                    )
            # Reset metrics or the data will keep accruing becoming an average of ALL the epochs
            self._reset_keras_metrics()

    @tf.function
    def train_step(
        self,
        x,
        adversarial_loss_weight: float,
        contextual_loss_weight: float,
        enc_loss_weight: float,
    ):

        # Random noise
        z = tf.random.normal((tf.shape(x)[0], self._hps["latent_vector_size"]))

        """Single training step."""
        with tf.GradientTape(persistent=True) as tape:
            # Generator reconstruction from random noise
            g_z = self.generator(z, training=True)  # or False?

            # Discriminator on real data
            d_x, _ = self.discriminator(x, training=True)

            # Reconstruct real data after encoding
            e_x = self.encoder(x, training=True)
            g_ex = self.generator(e_x, training=True)

            # Discriminator on the reconstructed real data g_ex
            d_gex, _ = self.discriminator(inputs=g_ex, training=True)

            # Encode the reconstructed real data g_ex
            e_gex = self.encoder(g_ex, training=True)

            # Discriminator Loss
            # d_loss = self._minmax(d_x_features, d_gex_features)
            d_loss = self._minmax(d_x, d_gex)

            # Generator Loss
            # adversarial_loss = losses.adversarial_loss_fm(d_f_x, d_f_x_hat)
            bce_g_loss = generator_bce(g_ex, from_logits=True)

            l1_loss = self._mae(x, g_ex)  # Contextual loss
            e_loss = self._mse(e_x, e_gex)  # Encoder loss

            g_loss = (
                adversarial_loss_weight * bce_g_loss
                + contextual_loss_weight * l1_loss
                + enc_loss_weight * e_loss
            )

        d_grads = tape.gradient(d_loss, self.discriminator.trainable_variables)
        g_grads = tape.gradient(g_loss, self.generator.trainable_variables)
        del tape

        self.optimizer_ge.apply_gradients(
            zip(
                g_grads,
                self.generator.trainable_variables + self.encoder.trainable_variables,
            )
        )
        self.optimizer_d.apply_gradients(
            zip(d_grads, self.discriminator.trainable_variables)
        )

        # NOTE: If d_loss = self._minmax(d_x_features, d_gex_features), dense layer would return
        # a warning. To suppress the warning "Gradients does not exist for variables" try using
        # the following lines.
        # self.optimizer_d.apply_gradients(
        #     (grad, var)
        #     for (grad, var) in zip(d_grads, self.discriminator.trainable_variables)
        #     if grad is not None
        # )

        return (
            g_z,
            g_ex,
            d_loss,
            g_loss,
            e_loss,
        )

    def test(self, base_path: Union[Path, None] = None):
        """
        Test the model on all the meaningful metrics.

        Args:
            base_path: the path to use for loading the models. If None, the default is used.
        """
        # Loop over every "best model" for every metric used in model selection
        for metric in ["auc_rc", "auc_roc"]:
            if not base_path:
                base_path = self._log_dir / "results" / metric
            encoder_path = base_path / "encoder"
            generator_path = base_path / "generator"

            # Load the best models to use as the model here
            encoder = k.models.load_model(encoder_path)
            encoder.summary()
            generator = k.models.load_model(generator_path)
            generator.summary()

            # Resetting the state of the AUPRC variable
            self._auc_rc.reset_states()
            tf.print("Using the best model selected via ", metric)
            # Test on the test dataset
            for batch in self._dataset.test:
                x, labels_test = batch
                # Get the anomaly scores
                anomaly_scores = self._compute_anomaly_scores(x, encoder, generator)
                self._auc_rc.update_state(labels_test, anomaly_scores)
                self._auc_roc.update_state(labels_test, anomaly_scores)

            auc_rc = self._auc_rc.result()
            auc_roc = self._auc_roc.result()

            tf.print("Get AUC-RC: ", auc_rc)
            tf.print("Get AUC-ROC: ", auc_roc)

            base_path = self._log_dir / "results" / metric
            result = {
                "auc_roc": {
                    "value": float(auc_roc),
                },
                "auc_rc": {
                    "value": float(auc_rc),
                },
            }
            # Write the file
            with open(base_path / "test.json", "w") as fp:
                json.dump(result, fp)

    def _compute_anomaly_scores(
        self, x: tf.Tensor, encoder: k.Model, generator: k.Model
    ) -> tf.Tensor:
        """
        Compute the anomaly scores as indicated in the GANomaly paper
        https://arxiv.org/abs/1805.06725.

        Args:
            x: The batch of data to use to calculate the anomaly scores.

        Returns:
            The anomaly scores on the input batch, [0, 1] normalized.
        """

        # Get the generator reconstruction of a decoded input data
        e_x = encoder(x, training=False)
        g_ex = generator(e_x, training=False)

        # Encode the generated g_ex
        e_gex = encoder(g_ex, training=False)

        # Get the anomaly scores
        normalized_anomaly_scores, _ = tf.linalg.normalize(
            tf.norm(
                self._flatten(tf.abs(e_x - e_gex)),
                axis=1,
                keepdims=False,
            )
        )

        return normalized_anomaly_scores



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trainer generic structure."""

from abc import abstractmethod
from pathlib import Path
from typing import Dict, Set

import tensorflow as tf

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset


class Trainer:
    """
    The Trainer represent the class to be used by all trainer objects.
    The class has the basic members that every trainer should have.
    """

    def __init__(
        self,
        dataset: AnomalyDetectionDataset,
        hps: Dict,
        summary_writer: tf.summary.SummaryWriter,
        log_dir: Path,
    ):
        """
        Trainer ctor.

        Args:
            dataset: The anomaly detection dataset.
            hps: instance of hyperparameters. Configure the trainer with this set of values.
                 NOTE: the hyperparameter supported and needed by the training procedure are
                 available in the `hyperparameters` property.
            summary_writer: The tf.summary.SummaryWriter object to keep track of the training
                            procedure.
            log_dir: the directory to use when saving something without using the summary writer.
        """
        self._log_dir = log_dir
        self._dataset = dataset
        self._hps = hps
        self._summary_writer = summary_writer
        self._keras_metrics = {}

    @property
    def keras_metrics(self) -> Dict[str, tf.keras.metrics.Metric]:
        """Return the dictionary of keras metrics to measure during the training."""
        return self._keras_metrics

    @keras_metrics.setter
    def keras_metrics(self, metrics: Dict[str, tf.keras.metrics.Metric]):
        self._keras_metrics = metrics

    @staticmethod
    @abstractmethod
    def hyperparameters() -> Set[str]:
        """List of the hyperparameters name used by the trainer."""
        raise NotImplementedError

    def _reset_keras_metrics(self) -> None:
        """
        Reset all the metrics.
        """
        for metric in self._keras_metrics.values():
            metric.reset_states()



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilities for HPS search."""

import itertools
import json
from pathlib import Path
from typing import Callable, List

import tensorflow as tf
from tensorboard.plugins.hparams import api as hp


def hparam_parser(
    hparams_path: Path, experiment: str, hyperparameters_names: List[str]
) -> List[hp.HParam]:
    """
    JSON hyperparameters file parser.

    Args:
        hparams_path: The path of the JSON hparams file.
        experiment: The name of the experiment.
        hyperparameters_names: The list of hyperparameters to be used.

    Returns:
        A list of hyperparameters.
    """
    with open(hparams_path, "r", encoding="utf-8") as fp:
        data = json.load(fp)

    # Get the experiment name
    experiment_data = data[experiment]

    # Fill the hparam list
    hps: List[hp.HParam] = []
    for hparam_name in hyperparameters_names:

        # Current parameter object
        try:
            current_param = experiment_data[hparam_name]
        except KeyError:
            if hparam_name in ("anomalous_label", "class_label"):
                continue
            raise KeyError(
                f"{hparam_name} key does not exist inside the {hparams_path} file"
            ) from None

        # Get the 'hp' attribute taking the correct "type" from the JSON object (usually the
        # type is 'Discrete')
        hp_attr = getattr(hp, current_param["type"])

        # Fill the list with the hparam name and the value of the correct type (e.g., 'Discrete')
        hps.append(
            hp.HParam(
                hparam_name,
                hp_attr(current_param["values"]),
            )
        )
    names = [param.name for param in hps]
    anomalous_label_present = "anomalous_label" in names
    class_label_present = "class_label" in names

    if not class_label_present and not anomalous_label_present:
        raise RuntimeError(
            "One among anomalous_label and class_label is required in hps."
        )
    if not class_label_present and anomalous_label_present:
        hps.append(hp.HParam("class_label", hp.Discrete([0])))
    if class_label_present and not anomalous_label_present:
        hps.append(hp.HParam("anomalous_label", hp.Discrete([0])))
    return hps


def grid_search(
    experiment_func: Callable,
    hps: List[hp.HParam],
    metrics: List[hp.Metric],
    log_dir: Path,
) -> None:
    """
    Perform a grid search over the values of the hyperparameters passed.

    Given a set of hyperparameters, this function firstly creates their combinatorial product
    and then iterates over the result each time calling the given experiment_func.

    Args:
        experiment_func: Callable, usually coding a single run of the experiment
        hps: List of hyperparameters, encoded as hp.HParam
        metrics: List of metrics to log, encoded as hp.Metric
        log_dir: Log directory where the tf.summary.SummaryWriter will save data
    """
    # Log the hps
    with tf.summary.create_file_writer(str(log_dir)).as_default():
        hp.hparams_config(hparams=hps, metrics=metrics)

    # Build all combinations of the HPS values
    hps_values = []
    for entry in hps:
        if isinstance(entry.domain, hp.Discrete):
            hps_values.append(entry.domain.values)
        # TODO: Add support for other type of hps
    hps_combinations = itertools.product(*hps_values)

    # Then for each set of HPS run an experiment
    for session_num, hps_values in enumerate(hps_combinations):
        hps_run = {entry.name: value for entry, value in zip(hps, hps_values)}
        run_name = f"run-{session_num}"
        print(f"--- Starting trial: {run_name} ---")
        print(hps_run)
        experiment_func(
            hps_run,
            str(log_dir / run_name),
        )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MNIST dataset, split to be used for anomaly detection."""

from functools import partial
from typing import Optional, Tuple, Union

import tensorflow as tf
import tensorflow_datasets as tfds

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset


class MNIST(AnomalyDetectionDataset):
    """MNIST dataset, split to be used for anomaly detection.
    Note:
        The label 1 is for the ANOMALOUS class.
        The label 0 is for the NORMAL class.
    """

    def __init__(self):
        super().__init__()
        (self._train_raw, self._test_raw), info = tfds.load(
            "mnist",
            split=["train", "test"],
            as_supervised=True,
            with_info=True,
        )

        self._num_classes = info.features["label"].num_classes

    def configure(
        self,
        batch_size: int,
        new_size: Tuple[int, int],
        anomalous_label: Union[int, str, None] = None,
        class_label: Union[int, str, None] = None,
        shuffle_buffer_size: int = 10000,
        cache: bool = True,
        drop_remainder: bool = True,
        output_range: Tuple[float, float] = (-1.0, 1.0),
    ) -> None:
        """Configure the dataset. This makes all the object properties valid (not None).
        Args:
            batch_size: The dataset batch size.
            new_size: (H,W) of the input image.
            anomalous_label: If the raw dataset contains label, all the elements with
                             "anomalous_label" are converted to element of
                             self.anomalous_label class.
            class_label: If the raw dataset contains different classes (each one
                         containing both positive and negative samples) we can select
                         only one class to focus on (e.g. a dataset of industrial
                         defects on industrial objects composed of transistors and
                         pills and we are interested only in transistors and not on pills).
            shuffle_buffer_size: Buffer size used during the tf.data.Dataset.shuffle call.
            cache: If True, cache the dataset.
            drop_remainder: If True, when the dataset size is not a multiple of the dataset size,
                            the last batch will be dropped.
            output_range: A Tuple (min, max) containing the output range to use
                          for the processed images.
        """

        pipeline = partial(
            self.pipeline,
            new_size=new_size,
            batch_size=batch_size,
            shuffle_buffer_size=shuffle_buffer_size,
            cache=cache,
            drop_remainder=drop_remainder,
            output_range=output_range,
        )

        pipeline_train = partial(pipeline, is_training=True)
        pipeline_test = partial(pipeline, is_training=False)
        is_anomalous = lambda _, label: tf.equal(label, anomalous_label)
        is_normal = lambda _, label: tf.not_equal(label, anomalous_label)

        # 60000 train images -> 6000 per class -> 600 per class in validation set
        # do not overlap wih train images -> 6000 - 600 per class in training set
        per_class_dataset = [
            self._train_raw.filter(lambda _, y: tf.equal(y, label))
            for label in tf.range(self._num_classes, dtype=tf.int64)
        ]

        validation_raw = per_class_dataset[0].take(600)
        train_raw = per_class_dataset[0].skip(600)
        for i in range(1, self._num_classes):
            validation_raw = validation_raw.concatenate(per_class_dataset[i].take(600))
            train_raw = train_raw.concatenate(per_class_dataset[i].skip(600))

        # Train-data
        self._train_anomalous = (
            train_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_train)
        )
        self._train_normal = (
            train_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_train)
        )
        self._train = train_raw.map(
            lambda x, label: (
                x,
                tf.cast(tf.equal(label, anomalous_label), tf.int32),
            )
        ).apply(pipeline_train)

        # Validation data
        self._validation_anomalous = (
            validation_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_test)
        )
        self._validation_normal = (
            validation_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_test)
        )
        self._validation = validation_raw.map(
            lambda x, label: (
                x,
                tf.cast(tf.equal(label, anomalous_label), tf.int32),
            )
        ).apply(pipeline_test)

        # Test-data
        self._test_anomalous = (
            self._test_raw.filter(is_anomalous)
            .map(lambda x, _: (x, self.anomalous_label))
            .apply(pipeline_test)
        )
        self._test_normal = (
            self._test_raw.filter(is_normal)
            .map(lambda x, _: (x, self.normal_label))
            .apply(pipeline_test)
        )

        # Complete dataset with positive and negatives
        def _to_binary(x, y):
            if tf.equal(y, anomalous_label):
                return (x, self.anomalous_label)
            return (x, self.normal_label)

        self._test = self._test_raw.map(_to_binary).apply(pipeline_test)



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trainer for the AnoGAN model."""

import json
from pathlib import Path
from typing import Dict, Set, Tuple, Union

import tensorflow as tf
import tensorflow.keras as k

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.models.anogan import Discriminator, Generator
from anomaly_toolbox.trainers.trainer import Trainer


def residual_image(x: tf.Tensor, g_z: tf.Tensor) -> tf.Tensor:
    """Residual image. The absolute value of the difference
    beteen x and g_z.
    Args:
        x: The input image.
        g_z: The generated image.
    Returns:
        The residual image.
    """
    return tf.math.abs(x - g_z)


def residual_loss(x: tf.Tensor, g_z: tf.Tensor) -> tf.Tensor:
    """Residual loss. The mean of the residual image.
    Args:
        x: The input image.
        g_z: The generated image.
    Returns:
        a scalar, the computed mean.
    """
    return tf.reduce_mean(residual_image(x, g_z))


class AdversarialLoss(k.losses.Loss):
    """The Min-Max loss, used to train the discriminator."""

    def __init__(self):
        super().__init__()
        self._bce = k.losses.BinaryCrossentropy(from_logits=True)

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        """Compute the 2 cross entropies and sum them.
        Args:
            y_true: MUST be D(x).
            y_pred: MUST be D(G(z)).
        Returns:
            bce(1, D(x)) + bce(0, D(G(z))
        """
        d_real = y_true
        d_gen = y_pred
        real_loss = self._bce(tf.ones_like(d_real), d_real)
        generated_loss = self._bce(tf.zeros_like(d_gen), d_gen)
        return real_loss + generated_loss


class AnoGAN(Trainer):
    """AnoGAN Trainer."""

    def __init__(
        self,
        dataset: AnomalyDetectionDataset,
        hps: Dict,
        summary_writer: tf.summary.SummaryWriter,
        log_dir: Path,
    ):
        """Initialize AnoGAN Trainer."""
        super().__init__(
            dataset=dataset, hps=hps, summary_writer=summary_writer, log_dir=log_dir
        )

        # Models
        self.discriminator = Discriminator(n_channels=dataset.channels)
        self.generator = Generator(
            n_channels=dataset.channels, input_dimension=hps["latent_vector_size"]
        )
        self._validate_models((28, 28, dataset.channels), hps["latent_vector_size"])

        # Optimizers
        self.optimizer_g = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.optimizer_d = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.optimizer_z = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )

        # Losses
        self._minmax = AdversarialLoss()
        self._bce = k.losses.BinaryCrossentropy(from_logits=True)

        # Metrics
        self.epoch_d_loss_avg = k.metrics.Mean(name="epoch_discriminator_loss")
        self.epoch_g_loss_avg = k.metrics.Mean(name="epoch_generator_loss")

        self._auc_roc = k.metrics.AUC(num_thresholds=500)

        self.keras_metrics = {
            metric.name: metric
            for metric in [self.epoch_d_loss_avg, self.epoch_g_loss_avg, self._auc_roc]
        }

        # Variables and constants
        self._z_gamma = tf.Variable(tf.zeros((hps["latent_vector_size"],)))
        self._lambda = tf.constant(0.1)

    @staticmethod
    def hyperparameters() -> Set[str]:
        """List of the hyperparameters name used by the trainer."""
        return {"learning_rate", "latent_vector_size"}

    def _validate_models(
        self, input_dimension: Tuple[int, int, int], latent_vector_size: int
    ):
        fake_latent_vector = (1, latent_vector_size)
        self.generator(tf.zeros(fake_latent_vector), training=False)
        self.generator.summary()

        fake_batch_size = (1, *input_dimension)
        self.discriminator(tf.zeros(fake_batch_size), training=False)
        self.discriminator.summary()

    def _select_and_save(self) -> None:
        """Saves the models (generator and discriminator) and the
        AUC thresholds and value.
        """
        current_auc = self._auc_roc.result()
        base_path = self._log_dir / "results" / "auc"
        self.discriminator.save(
            str(base_path / "discriminator"),
            overwrite=True,
            include_optimizer=False,
        )
        self.generator.save(
            str(base_path / "generator"),
            overwrite=True,
            include_optimizer=False,
        )

        with open(base_path / "validation.json", "w") as fp:
            json.dump(
                {
                    "value": float(current_auc),
                    "thresholds": self._auc_roc.thresholds,
                },
                fp,
            )

    @tf.function
    def train(
        self,
        epochs: tf.Tensor,
        step_log_frequency: tf.Tensor,
    ) -> None:
        """
        Train the model for the desired number of epochs.
        Calls the `train_step` function in loop.
        Also performs model selection on AUC using a subset of the test set.

        Args:
            epochs: The number of training epochs.
            step_log_frequency: Number of steps to use for logging on CLI and
                                tensorboard.
        """
        best_auc = -1.0
        for epoch in tf.range(epochs):
            for batch in self._dataset.train_normal:
                # Perform the train step
                x, _ = batch
                x_hat, d_loss, g_loss = self.train_step(x)

                # Update the losses metrics
                self.epoch_d_loss_avg.update_state(d_loss)
                self.epoch_g_loss_avg.update_state(g_loss)
                step = self.optimizer_d.iterations

                if tf.math.equal(tf.math.mod(step, step_log_frequency), 0):
                    with self._summary_writer.as_default():
                        tf.summary.scalar(
                            "learning_rate", self.optimizer_g.learning_rate, step=step
                        )
                        tf.summary.scalar(
                            "g_loss", self.epoch_g_loss_avg.result(), step=step
                        )
                        tf.summary.scalar(
                            "d_loss", self.epoch_d_loss_avg.result(), step=step
                        )

                        tf.summary.image("generated", x_hat, step=step)

                    tf.print(
                        "Step ",
                        step,
                        ": d_loss: ",
                        self.epoch_d_loss_avg.result(),
                        ", g_loss: ",
                        self.epoch_g_loss_avg.result(),
                        ", lr: ",
                        self.optimizer_g.learning_rate,
                    )
            tf.print("Epoch ", epoch, " completed.")

            # Reset the metrics at the end of every epoch
            self._reset_keras_metrics()

            # Model selection every model_selection epochs because the test phase is
            # terribly slow.
            model_selection = tf.constant(10)
            if tf.not_equal(tf.math.mod(epoch, model_selection), 0):
                continue

            # Model selection using a subset of the validation set (for speed reasons)
            # Keep "batches" number of batch of positives, them same for the negatives
            # then un-batch them, and process every element independently.
            batches = 1
            validation_subset = self._dataset.validation_normal.take(
                batches
            ).concatenate(self._dataset.validation_anomalous.take(batches))
            # We need to search for z, hence we do this 1 element at a time (slow!)
            validation_subset = validation_subset.unbatch().batch(1)

            step = self.optimizer_d.iterations
            for idx, sample in enumerate(validation_subset):
                x, y = sample
                # self._z_gamma should be the z value that's likely
                # to produce x (from what the generator knows)
                anomaly_score = self.latent_search(
                    x, self.generator, self.discriminator
                )
                self._auc_roc.update_state(
                    y_true=y, y_pred=tf.expand_dims(anomaly_score, axis=0)
                )
                with self._summary_writer.as_default():
                    g_z = self.generator(tf.expand_dims(self._z_gamma, axis=0))
                    tf.summary.image(
                        "test/inoutres",
                        tf.concat(
                            [x, g_z, residual_image(x, g_z)],
                            axis=2,
                        ),
                        step=step + idx,
                    )
            current_auc = self._auc_roc.result()
            with self._summary_writer.as_default():
                tf.summary.scalar("auc", current_auc, step=step)
                tf.print("Validation AUC: ", current_auc)

            if best_auc < current_auc:
                tf.py_function(self._select_and_save, [], [])
                best_auc = current_auc

    @tf.function
    def train_step(
        self,
        x: tf.Tensor,
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        """
        Single training step.

        Args:
            x: A batch of images.

        Returns:
            x_hat: A batch of generated images.
            d_loss: The discriminator loss.
            g_loss: The generator loss.
        """
        noise = tf.random.normal((tf.shape(x)[0], self._hps["latent_vector_size"]))
        with tf.GradientTape(persistent=True) as tape:
            x_hat = self.generator(noise, training=True)

            d_x, _ = self.discriminator(x, training=True)
            d_x_hat, _ = self.discriminator(x_hat, training=True)

            # Losses
            d_loss = self._minmax(d_x, d_x_hat)
            g_loss = self._bce(tf.ones_like(d_x_hat), d_x_hat)

        d_grads = tape.gradient(d_loss, self.discriminator.trainable_variables)
        g_grads = tape.gradient(g_loss, self.generator.trainable_variables)
        del tape

        self.optimizer_d.apply_gradients(
            zip(d_grads, self.discriminator.trainable_variables)
        )
        self.optimizer_g.apply_gradients(
            zip(g_grads, self.generator.trainable_variables)
        )

        return x_hat, d_loss, g_loss

    def latent_search(
        self,
        x: tf.Tensor,
        generator: k.Model,
        discriminator: k.Model,
        gamma: tf.Tensor = tf.constant(500),
    ) -> tf.Tensor:
        """
        Search in the latent space the z value that's likely to be mapped with the input image x.
        This step returns the value of the latent vector.
        NOTE: this is slow, since it performs gamma optimization steps
        to find the value of z.

        Args:
            x: Test image.
            generator: The generator model to be used.
            discriminator: The discriminator model to be used.
            gamma: Number of optimization steps.

        Returns:
            anomaly_score at the end of the gamma steps.
        """
        tf.print("Searching z with ", gamma, " opt steps...")

        @tf.function
        def opt_step():
            """Optimization steps that optimizes the value of self._z_gamma."""

            with tf.GradientTape(watch_accessed_variables=False) as tape:
                tape.watch(self._z_gamma)
                x_hat = generator(tf.expand_dims(self._z_gamma, axis=0))
                residual_score = residual_loss(x, x_hat)

                d_x, _ = discriminator(x, training=False)
                d_x_hat, _ = discriminator(x_hat, training=False)
                discrimination_score = self._minmax(d_x, d_x_hat)

                anomaly_score = (
                    1.0 - self._lambda
                ) * residual_score + self._lambda * discrimination_score

            # we want to minimize the anomaly score
            grads = tape.gradient(anomaly_score, [self._z_gamma])
            self.optimizer_z.apply_gradients(zip(grads, [self._z_gamma]))
            return anomaly_score

        self._z_gamma.assign(tf.zeros_like(self._z_gamma))
        for _ in tf.range(gamma):
            anomaly_score = opt_step()
        return anomaly_score

    def test(self, base_path: Union[Path, None] = None):
        """
        Test the model on the test dataset.
        VERY slow because we search for the optimal z optimizing for every new image of
        the test set.

        Args:
            base_path: the path to use for loading the models. If None, the default is used.

        Returns:
            None.
        """
        if not base_path:
            base_path = self._log_dir / "results" / "auc"
        generator_path = base_path / "generator"
        discriminator_path = base_path / "discriminator"

        # Load the best models to use as the model here
        generator = tf.keras.models.load_model(generator_path)
        generator.summary()
        discriminator = tf.keras.models.load_model(discriminator_path)
        discriminator.summary()

        # Resetting the state of the AUC variable
        self._auc_roc.reset_states()

        # We need to search for z, hence we do this 1 element at a time (slow!)
        test_subset = self._dataset.test.unbatch().batch(1)

        for sample in test_subset:
            x, y = sample
            # self._z_gamma should be the z value that's likely
            # to produce x (from what the generator knows)
            anomaly_score = self.latent_search(x, generator, discriminator)
            self._auc_roc.update_state(
                y_true=y, y_pred=tf.expand_dims(anomaly_score, axis=0)
            )

        current_auc = self._auc_roc.result()

        tf.print("Best AUC on test set: ", current_auc)
        result_json_path = base_path / "test.json"

        # Write the file
        with open(result_json_path, "w") as fp:
            json.dump(
                {
                    "auc_roc": {
                        "value": float(current_auc),
                        "thresholds": self._auc_roc.thresholds,
                    }
                },
                fp,
            )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorflow import keras as k

KERNEL_INITIALIZER = k.initializers.RandomNormal(mean=0.0, stddev=0.02)
ALMOST_ONE = k.initializers.RandomNormal(mean=1.0, stddev=0.02)


class Discriminator(k.Model):
    """
    Assemble the EGBAD BiGAN Discriminator as a :obj:`tf.keras.Model`.
    """

    def __init__(
        self,
        n_channels: int = 3,
        latent_space_dimension: int = 200,
        l2_penalty: float = 0.0,
    ):
        """
        Assemble the EGBAD BiGAN Discriminator as a :obj:`tf.keras.Model`.

        Args:
            n_channels: depth of the output image.
            latent_space_dimension: Dimension (along the depth) of the resulting
                                    encoded representation.
            l2_penalty: The penalty of the regularizer.

        Return:
            The assembled model.
        """

        super().__init__()

        input_dimension = (28, 28, n_channels)
        self._input_layer = k.layers.InputLayer(input_shape=input_dimension)
        self._input_encoded_layer = k.layers.InputLayer(
            input_shape=latent_space_dimension
        )
        self._concat = k.layers.Concatenate()

        # Backbone to elaborate the image shape data.
        self._backbone = k.Sequential(
            [
                self._input_layer,
                k.layers.Conv2D(
                    64,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    128,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    256,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    512,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    128,
                    (2, 2),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(1, 1),
                    padding="valid",
                    use_bias=False,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Flatten(),
            ]
        )

        # Part that elaborate the encoded data.
        self._encode_latent = k.Sequential(
            [
                k.layers.Dense(128, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.LeakyReLU(0.2),
            ]
        )

        # Work on concatenated data.
        self._features = k.Sequential(
            [
                k.layers.Dense(1024, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.LeakyReLU(0.2),
                k.layers.Flatten(),
            ]
        )

        # Output layer.
        self._output = k.layers.Dense(
            1,
            kernel_initializer=KERNEL_INITIALIZER,
            use_bias=False,
        )

    def call(self, inputs, training=True):
        x, z = inputs

        d_x = self._backbone(x)
        d_z = self._encode_latent(z)

        concat_input = self._concat([d_x, d_z])

        features = self._features(concat_input)
        d_xz = self._output(features)

        return d_xz, features


class Decoder(k.Sequential):
    """
    Assemble the EGBAD BiGAN Decoder as a :obj:`tf.keras.Model`.
    """

    def __init__(
        self,
        n_channels: int = 3,
        latent_space_dimension: int = 200,
        l2_penalty: float = 0.0,
    ):
        """
        Assemble the EGBAD BiGAN Decoder as a :obj:`tf.keras.Model`.

        Note:
            This is designed to work with any image size in a fully-convolutional way.

        Args:
            n_channels: Depth of the output image.
            latent_space_dimension: Dimension (along the depth) of the resulting
                                    encoded representation.
            l2_penalty: The penalty for the kernel regularizer.
        """

        super().__init__(
            [
                k.layers.Input(shape=latent_space_dimension),
                k.layers.Dense(512, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.Dense(7 * 7 * 128, kernel_initializer=KERNEL_INITIALIZER),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Reshape(target_shape=(7, 7, 128)),
                k.layers.Conv2DTranspose(
                    256,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    128,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2DTranspose(
                    n_channels,
                    (1, 1),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(1, 1),
                    padding="same",
                    use_bias=False,
                ),
                k.layers.Activation(k.activations.tanh),
            ]
        )


class Encoder(k.Sequential):
    """
    Assemble the EGBAD BiGAN Encoder as a :obj:`tf.keras.Model`.
    """

    def __init__(
        self,
        n_channels: int = 3,
        latent_space_dimension: int = 200,
        l2_penalty: float = 0.0,
    ):
        """
        Assemble the EGBAD BiGAN Encoder as a :obj:`tf.keras.Model`.

        Note:
            This is designed to work with any image size in a fully-convolutional way.

        Args:
            n_channels: Depth of the input image.
            latent_space_dimension: Dimension (along the depth) of the resulting
                                    encoded representation.
            l2_penalty: The penalty for the kernel regularizer.

        """

        super().__init__(
            [
                k.layers.Input(shape=(28, 28, n_channels)),
                k.layers.Conv2D(
                    64,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    128,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    256,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    512,
                    (4, 4),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(2, 2),
                    padding="same",
                    use_bias=False,
                    kernel_regularizer=k.regularizers.l2(l2_penalty),
                ),
                k.layers.BatchNormalization(
                    beta_initializer=ALMOST_ONE,
                    gamma_initializer=ALMOST_ONE,
                    momentum=0.1,
                    epsilon=1e-5,
                ),
                k.layers.LeakyReLU(0.2),
                k.layers.Conv2D(
                    latent_space_dimension,
                    (2, 2),
                    kernel_initializer=KERNEL_INITIALIZER,
                    strides=(1, 1),
                    padding="valid",
                    use_bias=False,
                ),  # 1
                k.layers.Flatten(),
            ]
        )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Corrupted MNIST dataset, split to be used for anomaly detection."""

import tensorflow_datasets as tfds

from .mnist import MNIST


class CorruptedMNIST(MNIST):
    """Corrupted MNIST dataset, splitted to be used for anomaly detection."""

    def __init__(self, corruption_type="shot_noise"):
        """Corrupted MNIST dataset, split to be used for anomaly detection.
        Args:
            corruption_type: one among the available corruptions as reported in
                             https://www.tensorflow.org/datasets/catalog/mnist_corrupted
        """
        super().__init__()
        (self._train, self._test), _ = tfds.load(
            f"mnist_corrupted/{corruption_type}",
            split=["train", "test"],
            as_supervised=True,
            with_info=True,
        )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trainer for the BiGAN model used in EGBAD."""

import json
from pathlib import Path
from typing import Dict, Set, Union

import tensorflow as tf
import tensorflow.keras as k

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.losses.egbad import (
    AdversarialLoss,
    encoder_bce,
    generator_bce,
    residual_loss,
)
from anomaly_toolbox.models.egbad import Decoder, Discriminator, Encoder
from anomaly_toolbox.trainers.trainer import Trainer


class EGBAD(Trainer):
    """EGBAD Trainer."""

    def __init__(
        self,
        dataset: AnomalyDetectionDataset,
        hps: Dict,
        summary_writer: tf.summary.SummaryWriter,
        log_dir: Path,
    ):
        """Initialize EGBAD-BiGAN Networks."""
        super().__init__(
            dataset=dataset, hps=hps, summary_writer=summary_writer, log_dir=log_dir
        )

        # # Models # #
        n_channels = dataset.channels

        # Decoder (aka Generator)
        self.generator = Decoder(n_channels, self._hps["latent_vector_size"], 0.2)
        self.generator.summary()

        # Encoder
        self.encoder = Encoder(
            n_channels,
            self._hps["latent_vector_size"],
            0.2,
        )
        self.encoder.summary()

        # Discriminator
        self.discriminator = Discriminator(n_channels, self._hps["latent_vector_size"])

        # Instantiate and define with correct input shape
        fake_batch_size = (1, 28, 28, n_channels)
        fake_latent_vector = (1, self._hps["latent_vector_size"])
        self.generator.call(tf.zeros(fake_latent_vector))
        self.encoder.call(tf.zeros(fake_batch_size))
        self.discriminator([tf.zeros(fake_batch_size), tf.zeros(fake_latent_vector)])
        self.discriminator.summary()

        # Optimizers
        self.optimizer_g = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.optimizer_d = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.optimizer_e = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )

        # Training Metrics
        self.epoch_d_loss_avg = k.metrics.Mean(name="epoch_discriminator_loss")
        self.epoch_g_loss_avg = k.metrics.Mean(name="epoch_generator_loss")
        self.epoch_e_loss_avg = k.metrics.Mean(name="epoch_encoder_loss")
        self._auc_rc = k.metrics.AUC(name="auc_rc", curve="PR", num_thresholds=500)
        self._auc_roc = k.metrics.AUC(name="auc_roc", curve="ROC", num_thresholds=500)

        self.keras_metrics = {
            metric.name: metric
            for metric in [
                self.epoch_d_loss_avg,
                self.epoch_g_loss_avg,
                self.epoch_e_loss_avg,
                self._auc_rc,
                self._auc_roc,
            ]
        }

        self._alpha = tf.constant(0.9)

        self._minmax = AdversarialLoss(from_logits=True)
        self._flatten = k.layers.Flatten()

    @staticmethod
    def hyperparameters() -> Set[str]:
        """List of the hyperparameters name used by the trainer."""
        return {"learning_rate", "latent_vector_size"}

    def train(
        self,
        epochs: int,
        step_log_frequency: int = 100,
    ):
        best_auc_rc, best_auc_roc = -1, -1
        for epoch in tf.range(epochs):

            for batch in self._dataset.train_normal:
                x, _ = batch

                # Perform the train step
                g_z, g_ex, d_loss, g_loss, e_loss = self.train_step(x)

                # Update the losses metrics
                self.epoch_d_loss_avg.update_state(d_loss)
                self.epoch_g_loss_avg.update_state(g_loss)
                self.epoch_e_loss_avg.update_state(e_loss)

                step = self.optimizer_d.iterations.numpy()
                learning_rate = self.optimizer_g.learning_rate.numpy()

                if tf.equal(tf.math.mod(step, step_log_frequency), 0):
                    with self._summary_writer.as_default():
                        tf.summary.scalar("learning_rate", learning_rate, step=step)
                        tf.summary.image(
                            "x/g_z/g_ex",
                            tf.concat([x, g_z, g_ex], axis=2),
                            step=step,
                        )
                        tf.summary.scalar(
                            "d_loss",
                            self.epoch_d_loss_avg.result(),
                            step=step,
                        )
                        tf.summary.scalar(
                            "g_loss",
                            self.epoch_g_loss_avg.result(),
                            step=step,
                        )
                        tf.summary.scalar(
                            "e_loss",
                            self.epoch_e_loss_avg.result(),
                            step=step,
                        )

                    tf.print(
                        "Step {:04d}: d_loss: {:.3f}, g_loss: {:.3f}, e_loss: {:.3f}, "
                        "lr: {:.5f}".format(
                            step,
                            self.epoch_d_loss_avg.result(),
                            self.epoch_g_loss_avg.result(),
                            self.epoch_e_loss_avg.result(),
                            learning_rate,
                        )
                    )

            # Epoch end
            tf.print(epoch, "Epoch completed")

            # Model selection at the end of every epoch on the validation set
            # 1. Compute the the anomaly score
            # 2. Use the AUC object to compute the AUC with different
            # thresholds values on the anomaly score.
            self._auc_rc.reset_state()
            self._auc_roc.reset_state()

            for batch in self._dataset.validation:
                x, labels_test = batch

                # Get the anomaly scores
                anomaly_scores = self._compute_anomaly_scores(
                    x, self.encoder, self.generator, self.discriminator
                )

                # Update streaming auc_rc
                self._auc_rc.update_state(labels_test, anomaly_scores)
                # Update streaming auc_roc
                self._auc_roc.update_state(labels_test, anomaly_scores)

            # Save the best models (AUC RC)
            current_auc_rc = self._auc_rc.result()
            if best_auc_rc < current_auc_rc:
                tf.print("Best AUPRC on validation set: ", current_auc_rc)

                # Replace the best
                best_auc_rc = current_auc_rc

                base_path = self._log_dir / "results" / "auc_rc"
                self.generator.save(str(base_path / "generator"), overwrite=True)
                self.encoder.save(str(base_path / "encoder"), overwrite=True)
                self.discriminator.save(
                    str(base_path / "discriminator"), overwrite=True
                )

                with open(base_path / "validation.json", "w") as fp:
                    json.dump(
                        {
                            "value": float(best_auc_rc),
                        },
                        fp,
                    )

            # Save the best models (AUC ROC)
            current_auc_roc = self._auc_rc.result()
            if best_auc_roc < current_auc_roc:
                tf.print("Best AUC ROC on validation set: ", current_auc_roc)

                # Replace the best
                best_auc_rc = current_auc_roc

                base_path = self._log_dir / "results" / "auc_roc"
                self.generator.save(str(base_path / "generator"), overwrite=True)
                self.encoder.save(str(base_path / "encoder"), overwrite=True)
                self.discriminator.save(
                    str(base_path / "discriminator"), overwrite=True
                )

                with open(base_path / "validation.json", "w") as fp:
                    json.dump(
                        {
                            "value": float(best_auc_rc),
                        },
                        fp,
                    )

            # Reset metrics or the data will keep accruing becoming an average of ALL the epochs
            self._reset_keras_metrics()

    @tf.function
    def train_step(
        self,
        x,
    ):
        """Single training step."""
        z = tf.random.normal((tf.shape(x)[0], self._hps["latent_vector_size"]))
        with tf.GradientTape(persistent=True) as tape:
            # Reconstruction
            g_z = self.generator.call(z, training=True)

            e_x = self.encoder.call(x, training=True)
            g_ex = self.generator.call(e_x, training=True)

            d_g_z, _ = self.discriminator(inputs=[g_z, z], training=True)
            d_x, _ = self.discriminator(inputs=[x, e_x], training=True)

            # Losses
            d_loss = self._minmax(d_x, d_g_z)

            # g_loss = self._bce(tf.ones_like(d_g_z), d_g_z)
            g_loss = generator_bce(d_g_z, from_logits=True)

            # e_loss = self._bce(tf.zeros_like(d_x), d_x) + residual_loss(x, g_ex)
            e_loss = encoder_bce(d_x, from_logits=True) + residual_loss(x, g_ex)

        d_grads = tape.gradient(d_loss, self.discriminator.trainable_variables)
        g_grads = tape.gradient(g_loss, self.generator.trainable_variables)
        e_grads = tape.gradient(e_loss, self.encoder.trainable_variables)
        del tape

        self.optimizer_d.apply_gradients(
            zip(d_grads, self.discriminator.trainable_variables)
        )
        self.optimizer_g.apply_gradients(
            zip(g_grads, self.generator.trainable_variables)
        )
        self.optimizer_e.apply_gradients(zip(e_grads, self.encoder.trainable_variables))

        return (
            g_z,
            g_ex,
            d_loss,
            g_loss,
            e_loss,
        )

    def test(self, base_path: Union[Path, None] = None):
        """Measure the performance on the test set.

        Args:
            base_path: the path to use for loading the models. If None, the default is used.
        """

        for metric in ["auc_rc", "auc_roc"]:
            if not base_path:
                base_path = self._log_dir / "results" / metric
            encoder_path = base_path / "encoder"
            generator_path = base_path / "generator"
            discriminator_path = base_path / "discriminator"

            # Load the best models to use as the model here
            encoder = tf.keras.models.load_model(encoder_path)
            encoder.summary()
            generator = tf.keras.models.load_model(generator_path)
            generator.summary()
            discriminator = tf.keras.models.load_model(discriminator_path)
            discriminator.summary()

            # Resetting the state of the AUPRC variable
            self._auc_rc.reset_states()

            # Test on the test dataset
            for batch in self._dataset.test:
                x, labels_test = batch

                anomaly_scores = self._compute_anomaly_scores(
                    x, encoder, generator, discriminator
                )

                self._auc_rc.update_state(labels_test, anomaly_scores)
                self._auc_roc.update_state(labels_test, anomaly_scores)

            auc_rc = self._auc_rc.result()
            auc_roc = self._auc_roc.result()

            tf.print("Best AUPRC on test set: ", auc_rc)
            tf.print("Best AUCROC on test set: ", auc_roc)

            base_path = self._log_dir / "results" / metric

            # Create the resulta
            result = {
                "auc_rc": {
                    "value": float(auc_rc),
                },
                "auc_roc": {
                    "value": float(auc_roc),
                },
            }

            # Write the file
            with open(base_path / "test.json", "w") as fp:
                json.dump(result, fp)

    def _compute_anomaly_scores(
        self, x: tf.Tensor, encoder: k.Model, generator: k.Model, discriminator: k.Model
    ) -> tf.Tensor:
        """
        Compute the anomaly scores as indicated in the EGBAD paper
        https://arxiv.org/abs/1802.06222.

        Args:
            x: The batch of data to use to calculate the anomaly scores.

        Returns:
            The anomaly scores on the input batch, [0, 1] normalized.

        """
        e_x = encoder.call(x, training=False)
        g_ex = generator.call(e_x, training=False)

        _, x_features = discriminator([x, e_x], training=False)
        _, ex_features = discriminator([g_ex, e_x], training=False)

        g_score = tf.norm(
            k.layers.Flatten()(residual_loss(x, g_ex)), axis=1, keepdims=False
        )

        # Remove unused (i.e., 1-shaped) axis by squeezing
        ex_features = tf.squeeze(ex_features)
        x_features = tf.squeeze(x_features)

        d_score = tf.norm(
            self._flatten(x_features - ex_features), axis=1, keepdims=False
        )

        # Anomaly score should have a shape of (batch_size,)
        normalized_anomaly_scores, _ = tf.linalg.normalize(
            self._alpha * d_score + (1.0 - self._alpha) * g_score
        )

        return normalized_anomaly_scores



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""All GANomaly experiments."""

from pathlib import Path
from typing import Dict

import tensorflow as tf

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.experiments.experiment import Experiment
from anomaly_toolbox.hps import hparam_parser
from anomaly_toolbox.trainers.ganomaly import GANomaly


class GANomalyExperiment(Experiment):
    """
    GANomaly experiment.
    """

    def __init__(self, hparams_path: Path, log_dir: Path):
        super().__init__(hparams_path, log_dir)

        # Get the hyperparameters
        self._hps = hparam_parser(
            self._hparams_path,
            "ganomaly",
            list(self.hyperparameters().union(GANomaly.hyperparameters())),
        )

    def experiment(
        self, hps: Dict, log_dir: Path, dataset: AnomalyDetectionDataset
    ) -> None:
        """Experiment execution - architecture specific.

        Args:
            hps: Dictionary with the parameters to use for the current run.
            log_dir: Where to store the tensorboard logs.
            dataset: The dataset to use for model training and evaluation.
        """
        print("Running GANomaly experiment...")

        summary_writer = tf.summary.create_file_writer(str(log_dir))
        new_size = (32, 32)

        # Create the dataset with the requested sizes (requested by the model architecture)
        dataset.configure(
            anomalous_label=hps["anomalous_label"],
            class_label=hps["class_label"],
            batch_size=hps["batch_size"],
            new_size=new_size,
            shuffle_buffer_size=hps["shuffle_buffer_size"],
        )

        # Create the GANomaly trainer
        trainer = GANomaly(
            dataset=dataset,
            hps=hps,
            summary_writer=summary_writer,
            log_dir=log_dir,
        )

        # Train the GANomaly model
        trainer.train(
            epochs=hps["epochs"],
            adversarial_loss_weight=hps["adversarial_loss_weight"],
            contextual_loss_weight=hps["contextual_loss_weight"],
            enc_loss_weight=hps["enc_loss_weight"],
        )

        # Test on test dataset and put the results in the json file (the same file  used inside the
        # training for the model selection)
        trainer.test()



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trainer for the DeScarGAN model."""

import json
from pathlib import Path
from typing import Dict, Set, Tuple, Union

import tensorflow as tf
import tensorflow.keras as k

from anomaly_toolbox.datasets.dataset import AnomalyDetectionDataset
from anomaly_toolbox.models.descargan import Discriminator, Generator
from anomaly_toolbox.trainers.trainer import Trainer


class DeScarGAN(Trainer):
    """DeScarGAN Trainer."""

    def __init__(
        self,
        dataset: AnomalyDetectionDataset,
        hps: Dict,
        summary_writer: tf.summary.SummaryWriter,
        log_dir: Path,
    ):
        """Initialize DeScarGAN Trainer."""
        super().__init__(
            dataset, hps=hps, summary_writer=summary_writer, log_dir=log_dir
        )

        # Data info
        self._ill_label = dataset.anomalous_label
        self._healthy_label = dataset.normal_label

        # Models
        self.generator = Generator(
            ill_label=self._ill_label, n_channels=dataset.channels
        )
        self.discriminator = Discriminator(
            ill_label=self._ill_label, n_channels=dataset.channels
        )

        # Optimizers
        self.g_optimizer = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )
        self.d_optimizer = k.optimizers.Adam(
            learning_rate=hps["learning_rate"], beta_1=0.5, beta_2=0.999
        )

        # Parameters
        self._generator_training_steps = tf.constant(5, dtype=tf.int64)
        self._g_lambda_identity = tf.constant(50.0)
        self._g_lambda_reconstruction = tf.constant(50.0)
        self._g_lambda_fake = tf.constant(1.0)
        self._g_lambda_classification = tf.constant(1.0)

        self._d_lambda_gradient_penalty = tf.constant(10.0)
        self._d_lambda_fake = tf.constant(20.0)
        self._d_lambda_real = tf.constant(20.0)
        self._d_lambda_classification = tf.constant(5.0)

        # Losses
        self._classification_loss = k.losses.SparseCategoricalCrossentropy(
            from_logits=True
        )
        self._reconstruction_error = k.losses.MeanSquaredError()

        # Training Metrics
        self.epoch_d_loss_avg = k.metrics.Mean(name="epoch_discriminator_loss")
        self.epoch_g_loss_avg = k.metrics.Mean(name="epoch_generator_loss")
        self.accuracy = k.metrics.BinaryAccuracy(name="binary_accuracy")
        self.keras_metrics = {
            metric.name: metric
            for metric in [self.epoch_d_loss_avg, self.epoch_g_loss_avg, self.accuracy]
        }
        # Outside of the keras_metrics because it's used to compute the running
        # mean and not as a metric
        self._mean = k.metrics.Mean()

        # Constants
        self._zero = tf.constant(0.0)
        self._zero_batch = tf.zeros((1, 1, 1, 1))

    @staticmethod
    def hyperparameters() -> Set[str]:
        """List of the hyperparameters name used by the trainer."""
        return {"learning_rate"}

    @staticmethod
    def clip_by_norm_handle_none(grad, clip_norm):
        """
        tape.compute_gradients returns None instead of a zero tensor when the
        gradient would be a zero tensor and tf.clip_by_* does not support
        a None value.
        So just don't pass None to it and preserve None values.
        Source: https://stackoverflow.com/a/39295309/2891324
        """
        if grad is None:
            return None
        return tf.clip_by_norm(grad, clip_norm=clip_norm)

    def _select_and_save(self, threshold: tf.Tensor):
        current_accuracy = self.accuracy.result()
        base_path = self._log_dir / "results" / "accuracy"
        self.generator.save(
            str(base_path / "generator"),
            overwrite=True,
            include_optimizer=False,
        )
        with open(base_path / "validation.json", "w") as fp:
            json.dump(
                {
                    "value": float(current_accuracy),
                    "threshold": float(threshold),
                },
                fp,
            )

    @tf.function
    def train(
        self,
        epochs: int,
        step_log_frequency: int = 100,
    ):
        """
        Train the DeScarGAN generator and discriminator.
        """

        step_log_frequency = tf.convert_to_tensor(step_log_frequency, dtype=tf.int64)
        epochs = tf.convert_to_tensor(epochs, dtype=tf.int32)
        best_accuracy = -1.0
        for epoch in tf.range(epochs):
            for batch in self._dataset.train:
                # Perform the train step
                d_loss, g_loss, reconstructions = self.train_step(batch)

                # Update the losses metrics
                self.epoch_d_loss_avg.update_state(d_loss)
                self.epoch_g_loss_avg.update_state(g_loss)

                # step -1 because in train_step self.d_optimizer.iterations has been incremented
                if tf.math.equal(
                    tf.math.mod(self.d_optimizer.iterations - 1, step_log_frequency),
                    tf.constant(0, tf.int64),
                ):
                    x, y = batch

                    healthy_idx = tf.squeeze(tf.where(tf.equal(y, self._healthy_label)))
                    x_healthy = tf.gather(x, healthy_idx)
                    x_hat_healthy = tf.gather(reconstructions, healthy_idx)
                    if tf.equal(tf.rank(x_healthy), tf.constant(3)):
                        x_healthy = tf.expand_dims(x_healthy, axis=0)
                        x_hat_healthy = tf.expand_dims(x_hat_healthy, axis=0)

                    ill_idx = tf.squeeze(tf.where(tf.equal(y, self._ill_label)))
                    x_ill = tf.gather(x, ill_idx)
                    x_hat_ill = tf.gather(reconstructions, ill_idx)
                    if tf.equal(tf.rank(x_ill), tf.constant(3)):
                        x_ill = tf.expand_dims(x_ill, axis=0)
                        x_hat_ill = tf.expand_dims(x_hat_ill, axis=0)
                    with self._summary_writer.as_default():
                        tf.summary.scalar(
                            "d_loss", d_loss, step=self.d_optimizer.iterations
                        )
                        tf.summary.scalar(
                            "g_loss", g_loss, step=self.d_optimizer.iterations
                        )

                        tf.summary.image(
                            "healthy",
                            tf.concat(
                                [
                                    x_healthy,
                                    x_hat_healthy,
                                    tf.abs(x_healthy - x_hat_healthy),
                                ],
                                axis=2,
                            ),
                            step=self.d_optimizer.iterations,
                        )
                        tf.summary.image(
                            "ill",
                            tf.concat(
                                [x_ill, x_hat_ill, tf.abs(x_ill - x_hat_ill)], axis=2
                            ),
                            step=self.d_optimizer.iterations,
                        )
                    tf.print(
                        "[",
                        epoch,
                        "] step: ",
                        self.d_optimizer.iterations,
                        ": d_loss: ",
                        d_loss,
                        ", g_loss: ",
                        g_loss,
                    )
            # Epoch end

            # Global classification (not pixel level)
            # 1. Find the reconstruction error on a sub-set of the validation set
            # that WON'T be used during the score computation.
            # Reason: https://stats.stackexchange.com/a/427468/91290
            #
            # "The error distribution on the training data is misleading since your
            # training error distribution is not identical to test error distribution,
            # due to inevitable over-fitting. Then, comparing training error
            # distribution with future data is unjust."
            #
            # 2. Use the threshold to classify the validation set (positive and negative)
            # 3. Compute the binary accuracy (we can use it since the dataset is perfectly balanced)
            self._mean.reset_state()
            for x, y in self._dataset.validation_normal:
                self._mean.update_state(
                    tf.reduce_mean(
                        tf.math.abs(self.generator((x, y), training=False) - x)
                    )
                )
            threshold = self._mean.result()
            tf.print(
                "Reconstruction error on normal validation set: ",
                threshold,
            )

            # reconstruction <= threshold, then is a normal data (label 0)
            for x, y in self._dataset.test_normal.concatenate(
                self._dataset.test_anomalous
            ):
                self.accuracy.update_state(
                    y_true=y,
                    y_pred=tf.cast(
                        # reconstruction > threshold, then is anomalous (label 1 = cast(True))
                        # invoke the generator always with the normal label, since that's
                        # what we suppose to receive in input (and the threshold has been found
                        # using data that comes only from the normal distribution)
                        tf.math.greater(
                            tf.reduce_mean(
                                tf.math.abs(
                                    self.generator(
                                        (
                                            x,
                                            tf.ones(tf.shape(x)[0], dtype=tf.int32)
                                            * self._dataset.normal_label,
                                        ),
                                        training=False,
                                    )
                                    - x
                                ),
                                axis=[1, 2, 3],
                            ),
                            threshold,
                        ),
                        tf.int32,
                    ),
                )
            current_accuracy = self.accuracy.result()
            tf.print("Binary accuracy on validation set: ", current_accuracy)

            if best_accuracy < current_accuracy:
                tf.py_function(self._select_and_save, [threshold], [])
                best_accuracy = current_accuracy

            with self._summary_writer.as_default():
                tf.summary.scalar(
                    "accuracy", current_accuracy, step=self.d_optimizer.iterations
                )

            # Reset the metrics at the end of every epoch
            self._reset_keras_metrics()

    def gradient_penalty(
        self, x: tf.Tensor, x_gen: tf.Tensor, labels: tf.Tensor
    ) -> tf.Tensor:
        """
        Compute gradient penalty: L2(grad - 1)^2.

        Args:
            x: input batch
            x_gen: generated images
            labels: labels associated with x (and thus with x_gen)

        Returns:
            penalty on discriminator gradient
        """

        epsilon = tf.random.uniform([tf.shape(x)[0], 1, 1, 1], 0.0, 1.0)
        x_hat = epsilon * x + (1 - epsilon) * x_gen
        labels = tf.cast(labels, tf.float32)
        with tf.GradientTape() as tape:
            tape.watch([x_hat, labels])
            d_hat = self.discriminator([x_hat, labels], training=True)
        gradients = tape.gradient(d_hat, x_hat)
        ddx = tf.sqrt(tf.reduce_sum(gradients ** 2, axis=[1, 2]))
        d_regularizer = tf.reduce_mean((ddx - 1.0) ** 2)
        return d_regularizer

    @tf.function
    def train_step(
        self,
        inputs: Tuple[tf.Tensor, tf.Tensor],
    ):
        """
        Single training step.

        Args:
            inputs: a tuple  (x,y) containing the input samples (x) and their labels (y).
                    both x, and y, are batches.
                    If x is a batch of images the input shape is (batch_size, h, w, d).
                    The shape of y is always (batch_size,).

        Returns:
            d_loss, g_loss, x_hat, where

            d_loss: discriminator loss
            g_loss: generator loss
            x_hat: a tensor with the same shape of inputs[0] containing the reconstructions.
        """
        x, y = inputs

        # Generate all the reconstruction for the current batch
        # outside of the tape since we need this only for logging
        x_hat = self.generator(inputs, training=True)

        # All the gathering of the inputs can be done outside of the tape
        # no need to track these operations for computing the gradient (save memory)
        x_healthy = tf.gather(x, tf.squeeze(tf.where(tf.equal(y, self._healthy_label))))
        if tf.equal(tf.rank(x_healthy), tf.constant(3)):
            x_healthy = tf.expand_dims(x_healthy, axis=0)
        x_ill = tf.gather(x, tf.squeeze(tf.where(tf.equal(y, self._ill_label))))
        if tf.equal(tf.rank(x_ill), tf.constant(3)):
            x_ill = tf.expand_dims(x_ill, axis=0)

        # Count # healthy and # ill in batch
        tot_healthy = tf.cast(tf.shape(x_healthy)[0], tf.float32)
        tot_ill = tf.cast(tf.shape(x_ill)[0], tf.float32)
        tot = tf.cast(tf.shape(x)[0], tf.float32)
        percentage_healthy = tf.math.divide_no_nan(tot_healthy, tot)
        percentage_ill = tf.math.divide_no_nan(tot_ill, tot)

        # Scalar labels used in the losses
        healthy_labels = tf.ones((tot_healthy,), dtype=tf.int32) * tf.squeeze(
            self._healthy_label
        )
        ill_labels = tf.ones((tot_ill,), dtype=tf.int32) * tf.squeeze(self._ill_label)

        # Train the discriminator
        with tf.GradientTape(persistent=True) as tape:
            # With real images - healthy
            if tf.not_equal(percentage_healthy, self._zero):
                (d_healthy, d_healthy_pred) = self.discriminator(
                    [x_healthy, healthy_labels],
                    training=True,
                )
                d_loss_real_healthy = -tf.reduce_mean(d_healthy) * percentage_healthy
                d_loss_classification_healthy = (
                    self._classification_loss(
                        y_true=healthy_labels, y_pred=d_healthy_pred
                    )
                    * percentage_healthy
                )
            else:
                d_loss_classification_healthy = self._zero
                d_loss_real_healthy = self._zero

            # With real images - ill
            if tf.not_equal(percentage_ill, self._zero):
                (d_ill, d_ill_pred) = self.discriminator(
                    [x_ill, ill_labels], training=True
                )
                d_loss_real_ill = -tf.reduce_mean(d_ill) * percentage_ill
                d_loss_classification_ill = (
                    self._classification_loss(y_true=ill_labels, y_pred=d_ill_pred)
                    * percentage_ill
                )
            else:
                d_loss_classification_ill = self._zero
                d_loss_real_ill = self._zero

            # Total loss on real images
            d_loss_real = d_loss_real_ill + d_loss_real_healthy
            d_loss_classification = (
                d_loss_classification_ill + d_loss_classification_healthy
            )

            # Generate fake images:
            # Add random noise to the input too
            noise_variance = tf.constant(0.05)
            if tf.not_equal(percentage_healthy, self._zero):
                x_healthy_noisy = (
                    x_healthy
                    + tf.random.uniform(tf.shape(x_healthy), dtype=tf.float32)
                    * noise_variance
                )
                x_fake_healthy = self.generator(
                    [x_healthy_noisy, healthy_labels], training=True
                )
                # Add noise to generated and real images - used for the losses
                x_fake_healthy_noisy = (
                    x_fake_healthy
                    + tf.random.uniform(tf.shape(x_fake_healthy), dtype=tf.float32)
                    * noise_variance
                )
                # Train with fake noisy images
                (d_on_fake_healthy, _) = self.discriminator(
                    [x_fake_healthy_noisy, healthy_labels], training=True
                )

                # Gradient penealty
                d_gradient_penalty_healty = self.gradient_penalty(
                    x_healthy_noisy,
                    x_fake_healthy_noisy,
                    healthy_labels,
                )
            else:
                d_on_fake_healthy = self._zero
                d_gradient_penalty_healty = self._zero
                x_fake_healthy = self._zero
                x_fake_healthy_noisy = self._zero
                x_healthy_noisy = self._zero

            if tf.not_equal(percentage_ill, self._zero):
                x_ill_noisy = (
                    x_ill
                    + tf.random.uniform(tf.shape(x_ill), dtype=tf.float32)
                    * noise_variance
                )
                x_fake_ill = self.generator([x_ill_noisy, ill_labels], training=True)

                # Add noise to generated and real images - used for the losses
                x_fake_ill_noisy = (
                    x_fake_ill
                    + tf.random.uniform(tf.shape(x_fake_ill), dtype=tf.float32)
                    * noise_variance
                )

                # Train with fake noisy images
                (d_on_fake_ill, _) = self.discriminator(
                    [x_fake_ill_noisy, ill_labels], training=True
                )

                # Gradient penalty
                d_gradient_penalty_ill = self.gradient_penalty(
                    x_ill_noisy, x_fake_ill_noisy, ill_labels
                )
            else:
                d_on_fake_ill = self._zero_batch
                d_gradient_penalty_ill = self._zero
                x_fake_ill = self._zero_batch
                x_fake_ill_noisy = self._zero_batch
                x_ill_noisy = self._zero_batch

            d_loss_fake = (
                tf.reduce_mean(d_on_fake_healthy) * percentage_healthy
                + tf.reduce_mean(d_on_fake_ill) * percentage_ill
            )

            # Gradient penalty to improve discriminator training stability
            d_loss_gp = d_gradient_penalty_healty + d_gradient_penalty_ill

            # Sum all the losses and compute the discriminator loss
            d_loss = (
                self._d_lambda_real * d_loss_real
                + self._d_lambda_fake * d_loss_fake
                + self._d_lambda_classification * d_loss_classification
                + self._d_lambda_gradient_penalty * d_loss_gp
            )

            # Train the Generator ever self._generator_training_steps performed by the discriminator
            if tf.equal(
                tf.math.mod(
                    self.d_optimizer.iterations, self._generator_training_steps
                ),
                0,
            ):
                # D output reduction is needed because the output is batch_size, w, h, D

                if tf.not_equal(percentage_healthy, self._zero):
                    g_classification_loss_healthy = self._classification_loss(
                        y_true=healthy_labels,
                        y_pred=tf.reduce_mean(d_on_fake_healthy, axis=[2, 3]),
                    )
                    g_identity_loss_healthy = self._reconstruction_error(
                        y_true=x_healthy, y_pred=x_fake_healthy
                    )
                    g_reconstruction_loss_healthy = self._reconstruction_error(
                        y_true=x_healthy_noisy, y_pred=x_fake_healthy
                    )
                else:
                    g_classification_loss_healthy = self._zero
                    g_identity_loss_healthy = self._zero
                    g_reconstruction_loss_healthy = self._zero

                if tf.not_equal(percentage_ill, self._zero):
                    g_classification_loss_ill = self._classification_loss(
                        y_true=ill_labels,
                        y_pred=tf.reduce_mean(d_on_fake_ill, axis=[2, 3]),
                    )
                    g_identity_loss_ill = self._reconstruction_error(
                        y_true=x_ill, y_pred=x_fake_ill
                    )
                    g_reconstruction_loss_ill = self._reconstruction_error(
                        y_true=x_ill_noisy, y_pred=x_fake_ill
                    )
                else:
                    g_classification_loss_ill = self._zero
                    g_identity_loss_ill = self._zero
                    g_reconstruction_loss_ill = self._zero

                g_classification_loss = (
                    g_classification_loss_ill + g_classification_loss_healthy
                )

                # Adversarial loss
                g_loss_fake = -tf.reduce_mean(d_on_fake_healthy) - tf.reduce_mean(
                    d_on_fake_ill
                )

                # Identity loss
                g_identity_loss = g_identity_loss_ill + g_identity_loss_healthy

                # Reconstruction loss
                g_reconstruction_loss = (
                    g_reconstruction_loss_healthy + g_reconstruction_loss_ill
                )

                # Total generator loss
                g_loss = (
                    self._g_lambda_fake * g_loss_fake
                    + self._g_lambda_reconstruction * g_reconstruction_loss
                    + self._g_lambda_identity * g_identity_loss
                    + self._g_lambda_classification * g_classification_loss
                )
            else:
                g_loss = self._zero

        d_grads = tape.gradient(d_loss, self.discriminator.trainable_variables)
        # Gradient clipping
        d_grads = [self.clip_by_norm_handle_none(g, clip_norm=10) for g in d_grads]
        self.d_optimizer.apply_gradients(
            zip(d_grads, self.discriminator.trainable_variables)
        )

        if tf.equal(
            tf.cast(
                tf.math.mod(
                    self.d_optimizer.iterations - 1,
                    # -1 because at the previous line with d_opt.apply_gradients
                    # the counter increased
                    self._generator_training_steps,
                ),
                tf.int32,
            ),
            tf.constant(0, dtype=tf.int32),
        ):
            g_grads = tape.gradient(g_loss, self.generator.trainable_variables)
            # Gradient clipping
            g_grads = [self.clip_by_norm_handle_none(g, clip_norm=10) for g in g_grads]
            self.g_optimizer.apply_gradients(
                zip(g_grads, self.generator.trainable_variables)
            )
        del tape

        return d_loss, g_loss, x_hat

    def test(self, base_path: Union[Path, None] = None):
        """Measure the performance (only measured metric is accuracy) on the
        test set.

        Args:
            base_path: the path to use for loading the models. If None, the default is used.
        """

        if not base_path:
            base_path = self._log_dir / "results" / "accuracy"

        # Load the best model to use as the model here
        model_path = base_path / "generator"
        generator = tf.keras.models.load_model(model_path)
        generator.summary()

        self.accuracy.reset_state()

        # Get the threshold
        accuracy_path = base_path / "validation.json"
        with open(accuracy_path, "r") as fp:
            data = json.load(fp)
            threshold = data["threshold"]

        # reconstruction <= threshold => normal data (label 0)
        for x, y in self._dataset.test_normal.concatenate(self._dataset.test_anomalous):
            self.accuracy.update_state(
                y_true=y,
                y_pred=tf.cast(
                    # reconstruction > threshold => anomalous (label 1 = cast(True))
                    # invoke the generator always with the normal label, since that's
                    # what we suppose to receive in input (and the threshold has been found
                    # using data that comes only from the normal distribution)
                    tf.math.greater(
                        tf.reduce_mean(
                            tf.math.abs(
                                generator(
                                    (
                                        x,
                                        tf.ones(tf.shape(x)[0], dtype=tf.int32)
                                        * self._dataset.normal_label,
                                    ),
                                    training=False,
                                )
                                - x
                            ),
                            axis=[1, 2, 3],
                        ),
                        threshold,
                    ),
                    tf.int32,
                ),
            )

        current_accuracy = self.accuracy.result()
        tf.print("Binary accuracy on test set: ", current_accuracy)

        # Create the result
        result_path = self._log_dir / "results" / "accuracy" / "test.json"
        result_path.parent.mkdir(parents=True, exist_ok=True)

        # Write the file
        with open(result_path, "w") as fp:
            json.dump(
                {
                    "accuracy": {
                        "value": float(current_accuracy),
                        "threshold": float(threshold),
                    }
                },
                fp,
            )



# Copyright 2021 Zuru Tech HK Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The setup script."""

import sys

from setuptools import find_packages, setup


def main():
    """Setup script."""

    with open("README.md") as readme_file:
        readme = readme_file.read()

    requirements = open("requirements.in").read().split()

    setup(
        author="Zuru Tech HK Limited, All rights reserved.",
        author_email="ml@zuru.tech",
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Programming Language :: Python :: 3.7",
        ],
        description="Anomaly Toolbox Powered by GANs",
        install_requires=requirements,
        python_requires=">=3.8",
        license="MIT license",
        long_description=readme,
        long_description_content_type="text/markdown",
        include_package_data=True,
        keywords="anomaly_toolbox",
        name="anomaly_toolbox",
        scripts=["bin/anomaly-box.py"],
        package_dir={"": "src"},
        packages=find_packages(where="src"),
        url="https://github.com/zurutech/anomaly-toolbox",
        version="0.1.1",
        zip_safe=False,
    )


if __name__ == "__main__":
    sys.exit(main())
